import datetime
import os
import re
import time
from email.utils import formatdate
from hashlib import sha1
from itertools import chain
from typing import (
    Callable,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Union,
    overload,
)
from urllib.parse import quote

try:
    from mypy_extensions import mypyc_attr, trait
except ImportError:  # pragma: no cover

    def trait(cls):  # type: ignore
        return cls

    def mypyc_attr(*attrs, **kwattrs):  # type: ignore
        return lambda x: x


from .datastructures import Cookie, MutableHeaders
from .exceptions import MalformedRangeHeader, RangeNotSatisfiable
from .typing import Literal, ServerSentEvent


@mypyc_attr(allow_interpreted_subclasses=True)
class BaseResponse:
    def __init__(
        self, status_code: int = 200, headers: Optional[Mapping[str, str]] = None
    ) -> None:
        self.status_code = status_code
        self.headers = MutableHeaders(headers)
        self.cookies: List[Cookie] = []

    def set_cookie(
        self,
        key: str,
        value: str = "",
        max_age: int = -1,
        expires: Optional[int] = None,
        path: str = "/",
        domain: Optional[str] = None,
        secure: bool = False,
        httponly: bool = False,
        samesite: Literal["strict", "lax", "none"] = "lax",
    ) -> None:
        expires_datetime: Optional[datetime.datetime] = None
        if expires is not None:
            expires_datetime = datetime.datetime.fromtimestamp(time.time() + expires)

        self.cookies.append(
            Cookie(
                key,
                value,
                expires=expires_datetime,
                max_age=max_age,
                path=path,
                domain=domain,
                secure=secure,
                httponly=httponly,
                samesite=samesite,
            )
        )

    def delete_cookie(
        self,
        key: str,
        value: str = "",
        path: str = "/",
        domain: Optional[str] = None,
        secure: bool = False,
        httponly: bool = False,
        samesite: Literal["strict", "lax", "none"] = "lax",
    ) -> None:
        self.set_cookie(
            key,
            expires=0,
            max_age=0,
            path=path,
            domain=domain,
            secure=secure,
            httponly=httponly,
            samesite=samesite,
        )

    @overload
    def list_headers(self, *, as_bytes: Literal[True]) -> List[Tuple[bytes, bytes]]: ...

    @overload
    def list_headers(self, *, as_bytes: Literal[False]) -> List[Tuple[str, str]]: ...

    def list_headers(self, *, as_bytes):
        """
        Merge `self.headers` and `self.cookies` then returned as a list.
        """
        if as_bytes:
            return [
                *(
                    (key.encode("latin-1"), value.encode("latin-1"))
                    for key, value in self.headers.items()
                ),
                *((b"set-cookie", bytes(cookie)) for cookie in self.cookies),
            ]
        else:
            return [
                *self.headers.items(),
                *(("set-cookie", str(cookie)) for cookie in self.cookies),
            ]


@trait
class FileResponseMixin:
    def generate_common_headers(
        self,
        filepath: str,
        content_type: str,
        download_name: Optional[str],
        stat_result: os.stat_result,
    ) -> Dict[str, str]:
        headers: Dict[str, str] = {
            "accept-ranges": "bytes",
            "last-modified": formatdate(stat_result.st_mtime, usegmt=True),
            "etag": f'"{self.generate_etag(stat_result)}"',
        }
        if download_name or content_type == "application/octet-stream":
            download_name = download_name or os.path.basename(filepath)
            content_disposition = (
                "attachment; "
                f'filename="{download_name}"; '
                f"filename*=utf-8''{quote(download_name)}"
            )
            headers["content-disposition"] = content_disposition

        return headers

    @staticmethod
    def generate_etag(stat_result: os.stat_result) -> str:
        data = f"{stat_result.st_mtime}-{stat_result.st_size}"
        return sha1(data.encode("ascii")).hexdigest()

    @classmethod
    def judge_if_range(
        cls, if_range_raw_line: str, stat_result: os.stat_result
    ) -> bool:
        """
        Judge whether if-range is consistent with the value of etag or last-modified
        """
        return (
            if_range_raw_line == f'"{cls.generate_etag(stat_result)}"'
        ) or if_range_raw_line == formatdate(stat_result.st_mtime, usegmt=True)

    @staticmethod
    def parse_range(
        range_raw_line: str, max_size: int
    ) -> Sequence[Tuple[int, Union[int, int]]]:
        """
        Parse the Range header and make appropriate merge or cut processing
        """
        try:
            unit, ranges_str = range_raw_line.split("=", maxsplit=1)
        except ValueError:
            raise MalformedRangeHeader()
        if unit != "bytes":
            raise MalformedRangeHeader("Only support bytes range")

        ranges = [
            (
                int(_[0]) if _[0] else max_size - int(_[1]),
                int(_[1]) + 1 if _[0] and _[1] and int(_[1]) < max_size else max_size,
            )
            for _ in re.findall(r"(\d*)-(\d*)", ranges_str)
            if _ != ("", "")
        ]

        if len(ranges) == 0:
            raise MalformedRangeHeader("Range header: range must be requested")

        if any(not (0 <= start < max_size) for start, _ in ranges):
            raise RangeNotSatisfiable(max_size)

        if any(start > end for start, end in ranges):
            raise MalformedRangeHeader("Range header: start must be less than end")

        if len(ranges) == 1:
            return ranges

        result: List[Tuple[int, int]] = []
        for start, end in ranges:
            for p in range(len(result)):
                p_start, p_end = result[p]
                if start > p_end:
                    continue
                elif end < p_start:
                    result.insert(p, (start, end))
                    break
                else:
                    result[p] = (min(start, p_start), max(end, p_end))
                    break
            else:
                result.append((start, end))
        return result

    def generate_multipart(
        self,
        ranges: Sequence[Tuple[int, int]],
        boundary: str,
        max_size: int,
        content_type: str,
    ) -> Tuple[int, Callable[[int, int], bytes]]:
        r"""
        Multipart response headers generator.

        ```
        --{boundary}\n
        Content-Type: {content_type}\n
        Content-Range: bytes {start}-{end-1}/{max_size}\n
        \n
        ..........content...........\n
        --{boundary}\n
        Content-Type: {content_type}\n
        Content-Range: bytes {start}-{end-1}/{max_size}\n
        \n
        ..........content...........\n
        --{boundary}--\n
        ```
        """
        boundary_len = len(boundary)
        static_header_part_len = (
            44 + boundary_len + len(content_type) + len(str(max_size))
        )
        content_length = sum(
            (len(str(start)) + len(str(end - 1)) + static_header_part_len)  # Headers
            + (end - start)  # Content
            for start, end in ranges
        ) + (
            5 + boundary_len  # --boundary--\n
        )
        return (
            content_length,
            lambda start, end: (
                f"--{boundary}\n"
                f"Content-Type: {content_type}\n"
                f"Content-Range: bytes {start}-{end - 1}/{max_size}\n"
                "\n"
            ).encode("latin-1"),
        )


def build_bytes_from_sse(event: ServerSentEvent, charset: str) -> bytes:
    """
    helper function for SendEventResponse
    """
    data: Iterable[bytes]
    if "data" in event:
        data = (f"data: {_}".encode(charset) for _ in event.pop("data").splitlines())
    else:
        data = ()
    return b"\n".join(
        chain(
            map(lambda k, v: f"{k}: {v}".encode(charset), event.keys(), event.values()),
            data,
            (b"", b""),  # for generate b"\n\n"
        )
    )


def iri_to_uri(iri: str) -> str:
    """
    Convert an Internationalized Resource Identifier (IRI) portion to a URI portion
    that is suitable for inclusion in a URL.
    """
    # Copy from django
    # https://github.com/django/django/blob/main/django/utils/encoding.py#L100
    return quote(iri, safe="/#%[]=:;$&()+,!?*@'~")



import asyncio
import functools
from concurrent.futures import ThreadPoolExecutor as _ThreadPoolExecutor
from contextvars import copy_context
from typing import Any, Callable, TypeVar, cast

__all__ = [
    "ThreadPoolExecutor",
    "run_in_threadpool",
]


T = TypeVar("T")


class ThreadPoolExecutor(_ThreadPoolExecutor):  # type: ignore
    """
    Thread pool with ContextVars

    - https://github.com/python/cpython/issues/78195
    """

    def submit(self, __fn, *args, **kwargs):
        return super().submit(
            functools.partial(copy_context().run, __fn), *args, **kwargs
        )


async def run_in_threadpool(__fn: Callable[..., T], *args: Any, **kwargs: Any) -> T:
    """
    Asynchronously run function fn in a separate thread.
    Any *args and **kwargs supplied for this function are directly passed
    to fn. Also, the current :class:`contextvars.Context` is propogated,
    allowing context variables from the main thread to be accessed in the
    separate thread.

    https://github.com/python/cpython/blob/0f56263e62ba91d0baae40fb98947a3a98034a73/Lib/asyncio/threads.py
    """
    loop = asyncio.get_running_loop()
    ctx = copy_context()
    func_call = functools.partial(ctx.run, __fn, *args, **kwargs)
    return await loop.run_in_executor(None, cast(Callable[[], T], func_call))



import abc
import json
import os
import queue
import stat
from http import HTTPStatus
from mimetypes import guess_type
from random import choices as random_choices
from typing import (
    Any,
    Dict,
    Generator,
    Generic,
    Iterable,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
)

from baize.concurrency import ThreadPoolExecutor
from baize.datastructures import URL, defaultdict
from baize.exceptions import MalformedRangeHeader, RangeNotSatisfiable
from baize.responses import (
    BaseResponse,
    FileResponseMixin,
    build_bytes_from_sse,
    iri_to_uri,
)
from baize.typing import Environ, ServerSentEvent, StartResponse

StatusStringMapping = defaultdict(
    lambda status: f"{status} Unknown Status Code",
    {int(status): f"{status} {status.phrase}" for status in HTTPStatus},
)


class Response(BaseResponse):
    """
    The parent class of all responses, whose objects can be used directly as WSGI
    application.
    """

    def __call__(
        self, environ: Environ, start_response: StartResponse
    ) -> Iterable[bytes]:
        self.headers["content-length"] = "0"
        start_response(
            StatusStringMapping[self.status_code], self.list_headers(as_bytes=False)
        )
        return (b"",)


_ContentType = TypeVar("_ContentType")


class SmallResponse(Response, abc.ABC, Generic[_ContentType]):
    """
    Abstract base class for small response objects.
    """

    media_type: str = ""
    charset: str = "utf-8"

    def __init__(
        self,
        content: _ContentType,
        status_code: int = 200,
        headers: Optional[Mapping[str, str]] = None,
        media_type: Optional[str] = None,
        charset: Optional[str] = None,
    ) -> None:
        super().__init__(status_code, headers)
        self.content = content
        self.media_type = media_type or self.media_type
        self.charset = charset or self.charset

    @abc.abstractmethod
    def render(self, content: _ContentType) -> bytes:
        raise NotImplementedError

    def __call__(
        self, environ: Environ, start_response: StartResponse
    ) -> Iterable[bytes]:
        body = self.render(self.content)
        if body and "content-length" not in self.headers:
            content_length = str(len(body))
            self.headers["content-length"] = content_length
        content_type = self.media_type
        if content_type and "content-type" not in self.headers:
            if content_type.startswith("text/"):
                content_type += "; charset=" + self.charset
            self.headers["content-type"] = content_type
        start_response(
            StatusStringMapping[self.status_code], self.list_headers(as_bytes=False)
        )
        yield body


class PlainTextResponse(SmallResponse[Union[bytes, str]]):
    media_type = "text/plain"

    def render(self, content: Union[bytes, str]) -> bytes:
        return content.encode(self.charset) if isinstance(content, str) else content


class HTMLResponse(PlainTextResponse):
    media_type = "text/html"


class JSONResponse(SmallResponse[Any]):
    """
    `**kwargs` is used to accept all the parameters that `json.loads` can accept.
    """

    media_type = "application/json"

    def __init__(
        self,
        content: Any,
        status_code: int = 200,
        headers: Optional[Mapping[str, str]] = None,
        **kwargs: Any,
    ) -> None:
        self.json_kwargs: Dict[str, Any] = {
            "ensure_ascii": False,
            "allow_nan": False,
            "indent": None,
            "separators": (",", ":"),
            "default": None,
        }
        self.json_kwargs.update(**kwargs)
        super().__init__(content, status_code=status_code, headers=headers)

    def render(self, content: Any) -> bytes:
        return json.dumps(content, **self.json_kwargs).encode(self.charset)


class RedirectResponse(Response):
    def __init__(
        self,
        url: Union[str, URL],
        status_code: int = 307,
        headers: Optional[Mapping[str, str]] = None,
    ) -> None:
        super().__init__(status_code=status_code, headers=headers)
        self.headers["location"] = iri_to_uri(str(url))


class StreamingResponse(Response, abc.ABC, Generic[_ContentType]):
    def __init__(
        self,
        iterable: Iterable[_ContentType],
        status_code: int = 200,
        headers: Optional[Mapping[str, str]] = None,
    ) -> None:
        super().__init__(status_code, headers)
        self.iterable = iterable
        self._client_closed = False

    @abc.abstractmethod
    def render_stream(self) -> Generator[bytes, None, None]:
        raise NotImplementedError
        yield

    def __call__(
        self, environ: Environ, start_response: StartResponse
    ) -> Iterable[bytes]:
        start_response(
            StatusStringMapping[self.status_code], self.list_headers(as_bytes=False)
        )

        yield from self.render_stream()


class StreamResponse(StreamingResponse[bytes]):
    def __init__(
        self,
        iterable: Iterable[bytes],
        status_code: int = 200,
        headers: Optional[Mapping[str, str]] = None,
        content_type: str = "application/octet-stream",
    ) -> None:
        super().__init__(iterable, status_code, headers)
        self.headers["Content-Type"] = content_type

    def render_stream(self) -> Generator[bytes, None, None]:
        yield from self.iterable


class SendEventResponse(StreamingResponse[ServerSentEvent]):
    """
    Server-sent events response.

    When the cilent closes the connection, the generator will be closed.
    Use `try-finally` to clean up resources.

    ```python
    def generator():
        try:
            while True:
                yield ServerSentEvent()
        finally:
            print("generator closed")

    response = SendEventResponse(generator())
    ```
    """

    thread_pool = ThreadPoolExecutor(max_workers=10, thread_name_prefix="SendEvent_")

    required_headers = {
        "Cache-Control": "no-cache",
        # https://www.python.org/dev/peps/pep-3333/#other-http-features
        # WSGI application must not send connection header
        # "Connection": "keep-alive",
        "Content-Type": "text/event-stream",
    }

    def __init__(
        self,
        iterable: Iterable[ServerSentEvent],
        status_code: int = 200,
        headers: Optional[Mapping[str, str]] = None,
        *,
        ping_interval: float = 3,
        charset: str = "utf-8",
    ) -> None:
        if headers:
            headers = {**self.required_headers, **headers}
        else:
            headers = dict(self.required_headers)
        headers["Content-Type"] += f"; charset={charset}"
        super().__init__(iterable, status_code, headers)
        self.ping_interval = ping_interval
        self.charset = charset

    def render_stream(self) -> Generator[bytes, None, None]:
        q: "queue.Queue[ServerSentEvent | None]" = queue.Queue(maxsize=1)
        should_stop = False

        def push() -> None:
            nonlocal should_stop

            try:
                i = iter(self.iterable)

                while not should_stop:
                    try:
                        q.put(next(i))
                    except StopIteration:
                        should_stop = True
            finally:
                q.put(None)
                g = self.iterable
                if hasattr(g, "close"):
                    g.close()  # type: ignore

        push_future = self.thread_pool.submit(push)

        try:
            while not (push_future.done() and q.empty()):
                try:
                    event = q.get(timeout=self.ping_interval)
                    if event is None:
                        break
                    yield build_bytes_from_sse(event, self.charset)
                except queue.Empty:
                    yield b": ping\n\n"
        finally:
            should_stop = True
            while not q.empty():
                q.get_nowait()  # pragma: no cover
            if not push_future.cancel():
                exc = push_future.exception()
                if exc is not None:
                    raise exc


class FileResponse(Response, FileResponseMixin):
    """
    File response.

    It will automatically determine whether to send only headers
    and the range of files that need to be sent.
    """

    def __init__(
        self,
        filepath: str,
        headers: Optional[Mapping[str, str]] = None,
        content_type: Optional[str] = None,
        download_name: Optional[str] = None,
        stat_result: Optional[os.stat_result] = None,
        chunk_size: int = 4096 * 64,
    ) -> None:
        super().__init__(headers=headers)
        self.filepath = filepath
        self.content_type = (
            content_type
            or guess_type(download_name or os.path.basename(filepath))[0]
            or "application/octet-stream"
        )
        self.download_name = download_name
        self.stat_result = stat_result or os.stat(filepath)
        if stat.S_ISDIR(self.stat_result.st_mode):
            raise IsADirectoryError(f"{filepath} is a directory")
        self.chunk_size = chunk_size
        self.headers.update(
            self.generate_common_headers(
                self.filepath, self.content_type, self.download_name, self.stat_result
            )
        )

    def handle_all(
        self,
        send_header_only: bool,
        file_size: int,
        start_response: StartResponse,
    ) -> Generator[bytes, None, None]:
        self.headers["content-type"] = str(self.content_type)
        self.headers["content-length"] = str(file_size)
        start_response(StatusStringMapping[200], self.list_headers(as_bytes=False))

        if send_header_only:
            yield b""
            return

        with open(self.filepath, "rb") as file:
            for _ in range(0, file_size, self.chunk_size):
                yield file.read(self.chunk_size)

    def handle_single_range(
        self,
        send_header_only: bool,
        file_size: int,
        start_response: StartResponse,
        start: int,
        end: int,
    ) -> Generator[bytes, None, None]:
        self.headers["content-range"] = f"bytes {start}-{end - 1}/{file_size}"
        self.headers["content-type"] = str(self.content_type)
        self.headers["content-length"] = str(end - start)
        start_response(StatusStringMapping[206], self.list_headers(as_bytes=False))
        if send_header_only:
            yield b""
            return

        with open(self.filepath, "rb") as file:
            file.seek(start)
            for here in range(start, end, self.chunk_size):
                yield file.read(min(self.chunk_size, end - here))

    def handle_several_ranges(
        self,
        send_header_only: bool,
        file_size: int,
        start_response: StartResponse,
        ranges: Sequence[Tuple[int, int]],
    ) -> Generator[bytes, None, None]:
        boundary = "".join(random_choices("abcdefghijklmnopqrstuvwxyz0123456789", k=13))
        self.headers["content-type"] = f"multipart/byteranges; boundary={boundary}"
        content_length, generate_headers = self.generate_multipart(
            ranges, boundary, file_size, self.content_type
        )
        self.headers["content-length"] = str(content_length)

        start_response(StatusStringMapping[206], self.list_headers(as_bytes=False))
        if send_header_only:
            yield b""
            return

        with open(self.filepath, "rb") as file:
            for start, end in ranges:
                file.seek(start)
                yield generate_headers(start, end)
                for here in range(start, end, self.chunk_size):
                    yield file.read(min(self.chunk_size, end - here))
                yield b"\n"
            yield f"--{boundary}--\n".encode("ascii")

    def __call__(
        self, environ: Environ, start_response: StartResponse
    ) -> Iterable[bytes]:
        send_header_only = environ["REQUEST_METHOD"] == "HEAD"

        stat_result = self.stat_result
        file_size = stat_result.st_size

        if "HTTP_RANGE" not in environ or (
            "HTTP_IF_RANGE" in environ
            and not self.judge_if_range(environ["HTTP_IF_RANGE"], stat_result)
        ):
            yield from self.handle_all(send_header_only, file_size, start_response)
            return

        try:
            ranges = self.parse_range(environ["HTTP_RANGE"], file_size)
        except (MalformedRangeHeader, RangeNotSatisfiable) as exception:
            start_response(
                StatusStringMapping[exception.status_code],
                [*(exception.headers or {}).items()],
            )
            yield b"" if exception.content is None else exception.content.encode("utf8")
            return

        if len(ranges) == 1:
            start, end = ranges[0]
            yield from self.handle_single_range(
                send_header_only, file_size, start_response, start, end
            )
        else:
            yield from self.handle_several_ranges(
                send_header_only, file_size, start_response, ranges
            )



import asyncio
import json
from typing import Any, AsyncIterator, Dict, Iterator, Mapping
from urllib.parse import parse_qsl

from baize.datastructures import (
    URL,
    Address,
    FormData,
    Headers,
    QueryParams,
    UploadFile,
)
from baize.exceptions import MalformedJSON, MalformedMultipart, UnsupportedMediaType
from baize.multipart_helper import parse_async_stream as parse_multipart
from baize.requests import MoreInfoFromHeaderMixin
from baize.typing import Receive, Scope, Send
from baize.utils import cached_property

from .helper import empty_receive, empty_send


class ClientDisconnect(Exception):
    """
    HTTP connection disconnected.
    """


class HTTPConnection(Mapping[str, Any], MoreInfoFromHeaderMixin):
    """
    A base class for incoming HTTP connections.

    It is a valid Mapping type that allows you to directly
    access the values in any ASGI `scope` dictionary.
    """

    def __init__(
        self, scope: Scope, receive: Receive = empty_receive, send: Send = empty_send
    ) -> None:
        self._scope = scope
        self._send = send
        self._receive = receive

    def __getitem__(self, key: str) -> Any:
        return self._scope[key]

    def __iter__(self) -> Iterator[str]:
        return iter(self._scope)

    def __len__(self) -> int:
        return len(self._scope)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self._scope == other._scope
            and self._send == other._send
            and self._receive == other._receive
        )

    @cached_property
    def client(self) -> Address:
        """
        Client's IP and Port.

        Note that this depends on the "client" value given by
        the ASGI Server, and is not necessarily accurate.
        """
        host, port = self.get("client") or (None, None)
        return Address(host=host, port=port)

    @cached_property
    def url(self) -> URL:
        """
        The full URL of this request.
        """
        return URL(scope=self._scope)

    @cached_property
    def path_params(self) -> Dict[str, Any]:
        """
        The path parameters parsed by the framework.
        """
        return self.get("path_params", {})

    @cached_property
    def query_params(self) -> QueryParams:
        """
        Query parameter. It is a multi-value mapping.
        """
        return QueryParams(self["query_string"])

    @cached_property
    def headers(self) -> Headers:
        """
        A read-only case-independent mapping.

        Note that in its internal storage, all keys are in lower case.
        """
        return Headers(
            (key.decode("latin-1"), value.decode("latin-1"))
            for key, value in self._scope["headers"]
        )


class Request(HTTPConnection):
    def __init__(
        self, scope: Scope, receive: Receive = empty_receive, send: Send = empty_send
    ) -> None:
        assert scope["type"] == "http"
        super().__init__(scope, receive, send)
        self._stream_consumed = False
        self._is_disconnected = False

    @property
    def method(self) -> str:
        """
        HTTP method. Uppercase string.
        """
        return self["method"]

    async def stream(self) -> AsyncIterator[bytes]:
        """
        Streaming read request body. e.g. `async for chunk in request.stream(): ...`

        If you access `.stream()` then the byte chunks are provided
        without storing the entire body to memory. Any subsequent
        calls to `.body`, `.form`, or `.json` will raise an error.
        """
        if "body" in self.__dict__ and self.__dict__["body"].done():
            yield await self.body
            yield b""
            return

        if self._stream_consumed:
            raise RuntimeError("Stream consumed")

        self._stream_consumed = True
        while True:
            message = await self._receive()
            if message["type"] == "http.request":
                body = message.get("body", b"")
                if body:
                    yield body
                if not message.get("more_body", False):
                    break
            elif message["type"] == "http.disconnect":
                self._is_disconnected = True
                raise ClientDisconnect()
        yield b""

    @cached_property
    async def body(self) -> bytes:
        """
        Read all the contents of the request body into the memory and return it.
        """
        return b"".join([chunk async for chunk in self.stream()])

    @cached_property
    async def json(self) -> Any:
        """
        Call `await self.body` and use `json.loads` parse it.

        If `content_type` is not equal to `application/json`,
        an HTTPExcption exception will be thrown.
        """
        if self.content_type == "application/json":
            data = await self.body
            try:
                return json.loads(
                    data.decode(self.content_type.options.get("charset", "utf8"))
                )
            except json.JSONDecodeError as exc:
                raise MalformedJSON(str(exc)) from None

        raise UnsupportedMediaType("application/json")

    async def _parse_multipart(self, boundary: bytes, charset: str) -> FormData:
        return FormData(
            await parse_multipart(
                self.stream(), boundary, charset, file_factory=UploadFile
            )
        )

    @cached_property
    async def form(self) -> FormData:
        """
        Parse the data in the form format and return it as a multi-value mapping.

        If `content_type` is equal to `multipart/form-data`, it will directly
        perform streaming analysis, and subsequent calls to `self.body`
        or `self.json` will raise errors.

        If `content_type` is not equal to `multipart/form-data` or
        `application/x-www-form-urlencoded`, an HTTPExcption exception will be thrown.
        """
        if self.content_type == "multipart/form-data":
            charset = self.content_type.options.get("charset", "utf8")
            if "boundary" not in self.content_type.options:
                raise MalformedMultipart("Missing boundary in header content-type")
            boundary = self.content_type.options["boundary"].encode("latin-1")
            return await self._parse_multipart(boundary, charset)
        if self.content_type == "application/x-www-form-urlencoded":
            body = (await self.body).decode(
                encoding=self.content_type.options.get("charset", "latin-1")
            )
            return FormData(parse_qsl(body, keep_blank_values=True))

        raise UnsupportedMediaType(
            "multipart/form-data, application/x-www-form-urlencoded"
        )

    async def close(self) -> None:
        """
        Close all temporary files in the `self.form`.

        This can always be called, regardless of whether you use form or not.
        """
        if "form" in self.__dict__ and self.__dict__["form"].done():
            await (await self.form).aclose()

    async def is_disconnected(self) -> bool:
        """
        The method used to determine whether the connection is interrupted.
        """
        if not self._is_disconnected:
            try:
                message = await asyncio.wait_for(self._receive(), timeout=0.0000001)
                self._is_disconnected = message.get("type") == "http.disconnect"
            except asyncio.TimeoutError:
                pass
        return self._is_disconnected



import enum
from typing import AsyncIterator, Optional

from baize.typing import Message, Receive, Scope, Send

from .requests import HTTPConnection
from .responses import Response


class WebSocketDisconnect(Exception):
    def __init__(self, code: int = 1000, reason: Optional[str] = None) -> None:
        self.code = code
        self.reason = reason or ""

    def __repr__(self) -> str:
        return f"WebSocketDisconnect(code={self.code}, reason={self.reason})"


class WebSocketState(enum.Enum):
    CONNECTING = enum.auto()
    CONNECTED = enum.auto()
    DISCONNECTED = enum.auto()


class WebSocket(HTTPConnection):
    def __init__(self, scope: Scope, receive: Receive, send: Send) -> None:
        assert scope["type"] == "websocket"
        super().__init__(scope, receive, send)
        self.client_state = WebSocketState.CONNECTING
        self.application_state = WebSocketState.CONNECTING

    async def receive(self) -> Message:
        """
        Receive ASGI websocket messages, ensuring valid state transitions.
        """
        if self.client_state == WebSocketState.CONNECTING:
            message = await self._receive()
            message_type = message["type"]
            assert message_type == "websocket.connect"
            self.client_state = WebSocketState.CONNECTED
            return message
        elif self.client_state == WebSocketState.CONNECTED:
            message = await self._receive()
            message_type = message["type"]
            assert message_type in {"websocket.receive", "websocket.disconnect"}
            if message_type == "websocket.disconnect":
                self.client_state = WebSocketState.DISCONNECTED
            return message
        else:
            raise RuntimeError(
                'Cannot call "receive" once a disconnect message has been received.'
            )

    async def send(self, message: Message) -> None:
        """
        Send ASGI websocket messages, ensuring valid state transitions.
        """
        if self.application_state == WebSocketState.CONNECTING:
            message_type = message["type"]
            assert message_type in {"websocket.accept", "websocket.close"}
            if message_type == "websocket.close":
                self.application_state = WebSocketState.DISCONNECTED
            else:
                self.application_state = WebSocketState.CONNECTED
            await self._send(message)
        elif self.application_state == WebSocketState.CONNECTED:
            message_type = message["type"]
            assert message_type in {"websocket.send", "websocket.close"}
            if message_type == "websocket.close":
                self.application_state = WebSocketState.DISCONNECTED
            await self._send(message)
        else:
            raise RuntimeError('Cannot call "send" once a close message has been sent.')

    async def accept(self, subprotocol: Optional[str] = None) -> None:
        """
        Accept websocket connection.
        """
        if self.client_state == WebSocketState.CONNECTING:
            # If we haven't yet seen the 'connect' message, then wait for it first.
            await self.receive()
        await self.send({"type": "websocket.accept", "subprotocol": subprotocol})

    def _raise_on_disconnect(self, message: Message) -> None:
        if message["type"] == "websocket.disconnect":
            raise WebSocketDisconnect(message["code"], message.get("reason"))

    async def receive_text(self) -> str:
        """
        Receive a WebSocket text frame and return.
        """
        assert self.application_state == WebSocketState.CONNECTED
        message = await self.receive()
        self._raise_on_disconnect(message)
        return message["text"]

    async def receive_bytes(self) -> bytes:
        """
        Receive a WebSocket binary frame and return.
        """
        assert self.application_state == WebSocketState.CONNECTED
        message = await self.receive()
        self._raise_on_disconnect(message)
        return message["bytes"]

    async def iter_text(self) -> AsyncIterator[str]:
        """
        Keep receiving text frames until the WebSocket connection is disconnected.
        """
        try:
            while True:
                yield await self.receive_text()
        except WebSocketDisconnect:
            pass

    async def iter_bytes(self) -> AsyncIterator[bytes]:
        """
        Keep receiving binary frames until the WebSocket connection is disconnected.
        """
        try:
            while True:
                yield await self.receive_bytes()
        except WebSocketDisconnect:
            pass

    async def send_text(self, data: str) -> None:
        """
        Send a WebSocket text frame.
        """
        await self.send({"type": "websocket.send", "text": data})

    async def send_bytes(self, data: bytes) -> None:
        """
        Send a WebSocket binary frame.
        """
        await self.send({"type": "websocket.send", "bytes": data})

    async def close(self, code: int = 1000, reason: Optional[str] = None) -> None:
        """
        Close WebSocket connection. It can be called multiple times.
        """
        if self.application_state != WebSocketState.DISCONNECTED:
            await self.send({"type": "websocket.close", "code": code, "reason": reason})


WEBSOCKET_DENIAL_RESPONSE_MAPPING = {
    "http.response.start": "websocket.http.response.start",
    "http.response.body": "websocket.http.response.body",
}


class WebsocketDenialResponse:
    """
    A response that will deny a WebSocket connection.

    https://asgi.readthedocs.io/en/latest/extensions.html#websocket-denial-response
    """

    def __init__(self, response: Response) -> None:
        self.response = response

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        assert scope["type"] == "websocket", (
            "WebsocketDenialResponse requires a websocket scope"
        )

        # Check if Websocket Denial Response can be used
        if self.response is None or "websocket.http.response" not in scope.get(
            "extensions", {}
        ):
            await send({"type": "websocket.close"})
            return
        else:  # pragma: no cover
            # call the specified response, mapping send/receive events
            # between http/websocket ASGI protocols

            async def ws_send(msg: Message) -> None:
                if msg["type"] not in WEBSOCKET_DENIAL_RESPONSE_MAPPING:
                    raise ValueError(f"Unsupported message type: {msg['type']}")
                else:
                    msg["type"] = WEBSOCKET_DENIAL_RESPONSE_MAPPING[msg["type"]]
                await send(msg)

            async def ws_receive() -> Message:
                while True:
                    msg = await receive()
                    if msg["type"] == "websocket.disconnect":
                        msg["type"] = "http.disconnect"
                        return msg

            await self.response(scope, ws_receive, ws_send)



import abc
import re
import uuid
from datetime import date
from decimal import Decimal
from typing import Any, Dict, Generic, List, Optional, Pattern, Sequence, Tuple, TypeVar

try:
    from mypy_extensions import mypyc_attr
except ImportError:  # pragma: no cover

    def mypyc_attr(*attrs, **kwattrs):  # type: ignore
        return lambda x: x


from .typing import ASGIApp, WSGIApp

T = TypeVar("T")


@mypyc_attr(allow_interpreted_subclasses=True)
class Convertor(Generic[T]):
    regex: str

    @abc.abstractmethod
    def to_python(self, value: str) -> T:
        raise NotImplementedError()

    @abc.abstractmethod
    def to_string(self, value: T) -> str:
        raise NotImplementedError()


@mypyc_attr(allow_interpreted_subclasses=True)
class StringConvertor(Convertor[str]):
    regex = "[^/]+"

    def to_python(self, value: str) -> str:
        return value

    def to_string(self, value: str) -> str:
        value = str(value)
        if not value:
            raise ValueError("Must not be empty")
        if "/" in value:
            raise ValueError("May not contain path separators")
        return value


@mypyc_attr(allow_interpreted_subclasses=True)
class IntegerConvertor(Convertor[int]):
    regex = "[0-9]+"

    def to_python(self, value: str) -> int:
        return int(value)

    def to_string(self, value: int) -> str:
        if value < 0:
            raise ValueError("Negative integers are not supported")
        return str(value)


@mypyc_attr(allow_interpreted_subclasses=True)
class DecimalConvertor(Convertor[Decimal]):
    regex = "[0-9]+(.[0-9]+)?"

    def to_python(self, value: str) -> Decimal:
        return Decimal(value)

    def to_string(self, value: Decimal) -> str:
        if value.is_nan():
            raise ValueError("NaN values are not supported")
        if value.is_infinite():
            raise ValueError("Infinite values are not supported")
        if Decimal("0.0") > value:
            raise ValueError("Negative decimal are not supported")
        return str(value).rstrip("0").rstrip(".")


@mypyc_attr(allow_interpreted_subclasses=True)
class UUIDConvertor(Convertor[uuid.UUID]):
    regex = "[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}"

    def to_python(self, value: str) -> uuid.UUID:
        return uuid.UUID(value)

    def to_string(self, value: uuid.UUID) -> str:
        return str(value)


@mypyc_attr(allow_interpreted_subclasses=True)
class DateConvertor(Convertor[date]):
    regex = "[0-9]{4}-[0-9]{2}-[0-9]{2}"

    def to_python(self, value: str) -> date:
        return date(int(value[0:4]), int(value[5:7]), int(value[8:10]))

    def to_string(self, value: date) -> str:
        return value.isoformat()


@mypyc_attr(allow_interpreted_subclasses=True)
class AnyConvertor(Convertor[str]):
    regex = ".*"

    def to_python(self, value: str) -> str:
        return value

    def to_string(self, value: str) -> str:
        return value


CONVERTOR_TYPES: Dict[str, Convertor] = {
    "str": StringConvertor(),
    "int": IntegerConvertor(),
    "decimal": DecimalConvertor(),
    "uuid": UUIDConvertor(),
    "date": DateConvertor(),
    "any": AnyConvertor(),
}

# Match parameters in URL paths, eg. '{param}', and '{param:int}'
PARAM_REGEX = re.compile(r"{([^\d]\w*)(:\w+)?}")


def compile_path(path: str) -> Tuple[str, Dict[str, Convertor]]:
    """
    Given a path string, like: "/{username:str}", return a two-tuple
    of (format, {param_name:convertor}).

    format:     "/{username}"
    convertors: {"username": StringConvertor()}
    """
    path_format = ""
    idx = 0
    param_convertors = {}
    for match in PARAM_REGEX.finditer(path):
        param_name, convertor_type = match.groups("str")
        convertor_type = convertor_type.lstrip(":")
        if convertor_type not in CONVERTOR_TYPES:
            raise ValueError(f"Unknown path convertor '{convertor_type}'")
        convertor = CONVERTOR_TYPES[convertor_type]

        path_format += path[idx : match.start()]
        path_format += "{%s}" % param_name

        param_convertors[param_name] = convertor

        idx = match.end()

    path_format += path[idx:]

    return path_format, param_convertors


Interface = TypeVar("Interface", ASGIApp, WSGIApp)


@mypyc_attr(allow_interpreted_subclasses=True)
class Route(Generic[Interface]):
    def __init__(self, path: str, endpoint: Interface) -> None:
        self.path_format: str
        self.path_convertors: Dict[str, Convertor]
        self.path_format, self.path_convertors = compile_path(path)
        self.re_pattern = re.compile(
            self.path_format.format_map(
                {
                    name: f"(?P<{name}>{convertor.regex})"
                    for name, convertor in self.path_convertors.items()
                }
            )
        )
        self.endpoint: Interface = endpoint

    def matches(self, path: str) -> Tuple[bool, Dict[str, Any]]:
        match = self.re_pattern.fullmatch(path)
        if match is None:
            return False, {}
        return True, {
            name: self.path_convertors[name].to_python(value)
            for name, value in match.groupdict().items()
        }


@mypyc_attr(allow_interpreted_subclasses=True)
class BaseRouter(Generic[Interface]):
    def __init__(self, *routes: Tuple[str, Interface]) -> None:
        self._route_array: List[Route[Interface]] = [
            Route(path, endpoint) for path, endpoint in routes
        ]

    def search(self, path: str) -> Optional[Tuple[Route[Interface], Dict[str, Any]]]:
        for route in self._route_array:
            match_up, params = route.matches(path)
            if match_up:
                return route, params
        return None


@mypyc_attr(allow_interpreted_subclasses=True)
class BaseSubpaths(Generic[Interface]):
    def __init__(self, *routes: Tuple[str, Interface]) -> None:
        for prefix, _ in routes:
            if prefix == "":  # Allow use "" to set default app
                continue
            assert prefix.startswith("/"), "prefix must be starts with '/'"
            assert not prefix.endswith("/"), "prefix cannot be ends with '/'"
        self._route_array = [*routes]

    def search(self, path: str) -> Optional[Tuple[str, Interface]]:
        for prefix, endpoint in self._route_array:
            if path.startswith(prefix + "/") or path == prefix:
                return prefix, endpoint
        return None


@mypyc_attr(allow_interpreted_subclasses=True)
class BaseHosts(Generic[Interface]):
    def __init__(self, *hosts: Tuple[str, Interface]) -> None:
        self._host_array: Sequence[Tuple[Pattern, Interface]] = [
            (re.compile(host), endpoint) for host, endpoint in hosts
        ]

    def search(self, host: str) -> Optional[Interface]:
        for pattern, endpoint in self._host_array:
            if pattern.fullmatch(host) is not None:
                return endpoint
        return None



from baize.wsgi.requests import HTTPConnection, Request
from baize.wsgi.responses import (
    FileResponse,
    HTMLResponse,
    JSONResponse,
    PlainTextResponse,
    RedirectResponse,
    Response,
    SendEventResponse,
    SmallResponse,
    StreamResponse,
)
from baize.wsgi.routing import Hosts, Router, Subpaths
from baize.wsgi.shortcut import decorator, request_response
from baize.wsgi.staticfiles import Files, Pages
from baize.wsgi.middleware import NextRequest, NextResponse, middleware

__all__ = [
    "HTTPConnection",
    "Request",
    "Response",
    "SmallResponse",
    "PlainTextResponse",
    "HTMLResponse",
    "JSONResponse",
    "RedirectResponse",
    "StreamResponse",
    "FileResponse",
    "SendEventResponse",
    "Router",
    "Subpaths",
    "Hosts",
    "Files",
    "Pages",
    "request_response",
    "decorator",
    "NextRequest",
    "NextResponse",
    "middleware",
]



import importlib.util
import os
import stat
from email.utils import parsedate_to_datetime
from typing import Generic, Optional, Tuple, TypeVar, Union

from .responses import BaseResponse
from .typing import ASGIApp, Literal, WSGIApp

try:
    from mypy_extensions import mypyc_attr
except ImportError:  # pragma: no cover

    def mypyc_attr(*attrs, **kwattrs):  # type: ignore
        return lambda x: x


Interface = TypeVar("Interface", ASGIApp, WSGIApp)


@mypyc_attr(allow_interpreted_subclasses=True)
class BaseFiles(Generic[Interface]):
    def __init__(
        self,
        directory: Union[str, "os.PathLike[str]"],
        package: Optional[str] = None,
        *,
        handle_404: Optional[Interface] = None,
        cacheability: Literal["public", "private", "no-cache", "no-store"] = "public",
        max_age: int = 60 * 10,  # 10 minutes
    ) -> None:
        assert not (os.path.isabs(directory) and package is not None), (
            "directory must be a relative path, with package is not None"
        )
        self.directory = self.normalize_dir_path(str(directory), package)
        self.handle_404: Optional[Interface] = handle_404
        self.cacheability = cacheability
        self.max_age = max_age

    def normalize_dir_path(self, directory: str, package: Optional[str] = None) -> str:
        if package is None:
            return os.path.abspath(directory)
        else:
            spec = importlib.util.find_spec(package)
            assert spec is not None, f"Package {package!r} could not be found."
            assert spec.origin is not None, (
                f"Directory '{directory}' in package {package!r} could not be found."
            )
            package_directory = os.path.normpath(
                os.path.join(spec.origin, "..", directory)
            )
            assert os.path.isdir(package_directory), (
                f"Directory '{directory}' in package {package!r} could not be found."
            )
            return package_directory

    def ensure_absolute_path(self, path: str) -> Optional[str]:
        abspath = os.path.abspath(
            os.path.join(self.directory, os.path.join(*path.split("/")))
        )

        if path == "/":
            abspath += "/"

        if os.path.relpath(abspath, self.directory).startswith(".."):
            return None

        return abspath

    def check_path_is_file(
        self, path: Optional[str]
    ) -> Tuple[Optional[os.stat_result], bool]:
        if path is None:
            return None, False
        try:
            stat_result = os.stat(path)
            return stat_result, stat.S_ISREG(stat_result.st_mode)
        except FileNotFoundError:
            return None, False

    def if_none_match(self, etag: str, if_none_match: str) -> bool:
        if not if_none_match:
            return False

        if if_none_match == "*":
            return True

        if if_none_match.startswith("W/"):
            if_none_match = if_none_match[2:]

        return any(etag == i.strip().strip('"') for i in if_none_match.split(","))

    def if_modified_since(self, last_modified: float, if_modified_since: str) -> bool:
        try:
            if not if_modified_since:
                raise ValueError("Empty date value")
            modified_time = parsedate_to_datetime(if_modified_since).timestamp()
        except ValueError:
            return False

        return int(last_modified) <= int(modified_time)

    def set_response_headers(self, response: BaseResponse) -> None:
        response.headers.append(
            "Cache-Control", f"{self.cacheability}, max-age={self.max_age}"
        )
        response.headers.append("Vary", "Accept-Encoding, User-Agent, Cookie, Referer")



import functools
from typing import Any, Callable, Generator, Iterable, Iterator, MutableMapping, Tuple

from ..datastructures import Headers
from ..typing import Environ, StartResponse, WSGIApp
from .requests import Request
from .responses import Response, StreamingResponse


class NextRequest(Request, MutableMapping[str, Any]):
    def __setitem__(self, name: str, value: Any) -> None:
        self._environ[name] = value

    def __delitem__(self, name: str) -> None:
        del self._environ[name]

    def stream(self, chunk_size: int = 4096 * 16) -> Iterator[bytes]:
        raise RuntimeError("Cannot read request body in middleware.")


def ensure_next(iterable: Iterable[bytes]) -> Iterable[bytes]:
    first_chunk = iterable.__iter__().__next__()

    def generator():
        yield first_chunk
        yield from iterable

    return generator()


class NextResponse(StreamingResponse):
    """
    This is a response object for middleware.
    """

    def render_stream(self) -> Generator[bytes, None, None]:
        yield from self.iterable

    @classmethod
    def from_app(cls, app: WSGIApp, request: NextRequest) -> "NextResponse":
        """
        This is a helper method to convert a WSGI application into a NextResponse object.
        """
        status_code = 200
        headers: Headers = Headers()

        def start_response(
            status: str, response_headers: Iterable[Tuple[str, str]], exc_info=None
        ) -> None:
            nonlocal status_code
            nonlocal headers
            status_code = int(status.split(" ")[0])
            headers = Headers(response_headers)

        body = ensure_next(app(request, start_response))
        return NextResponse(body, status_code, headers)


def middleware(
    handler: Callable[[NextRequest, Callable[[NextRequest], NextResponse]], Response],
) -> Callable[[WSGIApp], WSGIApp]:
    """
    This can turn a callable object into a middleware for WSGI application.

    ```python
    @middleware
    def m(request: NextRequest, next_call: Callable[[NextRequest], NextResponse]) -> Response:
        ...
        response = next_call(request)
        ...
        return response

    @m
    @request_response
    def v(request: Request) -> Response:
        ...

    # OR

    @m
    def wsgi(environ: Environ, start_response: StartResponse) -> Iterable[bytes]:
        ...
    ```
    """

    @functools.wraps(handler)
    def d(app: WSGIApp) -> WSGIApp:
        """
        This is the actual middleware.
        """

        @functools.wraps(app)
        def wsgi(environ: Environ, start_response: StartResponse) -> Iterable[bytes]:
            request = NextRequest(environ, start_response)

            def next_call(request: NextRequest) -> NextResponse:
                next_response = NextResponse.from_app(app, request)
                return next_response

            response = handler(request, next_call)
            yield from response(environ, start_response)

        return wsgi

    return d



import os
import stat
from typing import Optional

from baize import staticfiles
from baize.datastructures import URL
from baize.exceptions import HTTPException
from baize.typing import ASGIApp, Receive, Scope, Send

from .responses import FileResponse, RedirectResponse, Response


class Files(staticfiles.BaseFiles[ASGIApp]):
    """
    Provide the ASGI application to download files in the specified path or
    the specified directory under the specified package.

    Support request range and cache (304 status code).
    """

    def file_response(
        self,
        filepath: str,
        stat_result: os.stat_result,
        if_none_match: str,
        if_modified_since: str,
    ) -> Response:
        if self.if_none_match(
            FileResponse.generate_etag(stat_result), if_none_match
        ) or self.if_modified_since(stat_result.st_mtime, if_modified_since):
            response = Response(304)
        else:
            response = FileResponse(filepath, stat_result=stat_result)
            self.set_response_headers(response)
        return response

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if_none_match: str = ""
        if_modified_since: str = ""
        for k, v in scope["headers"]:
            if k == b"if-none-match":
                if_none_match = v.decode("latin-1")
            elif k == b"if-modified-since":
                if_modified_since = v.decode("latin-1")
        filepath = self.ensure_absolute_path(scope["path"])
        stat_result, is_file = self.check_path_is_file(filepath)
        if is_file and stat_result:
            assert filepath is not None  # Just for type check
            return await self.file_response(
                filepath, stat_result, if_none_match, if_modified_since
            )(scope, receive, send)

        if self.handle_404 is None:
            raise HTTPException(404)
        else:
            return await self.handle_404(scope, receive, send)


class Pages(Files):
    """
    Provide the ASGI application to download files in the specified path or
    the specified directory under the specified package.
    Unlike `Files`, when you visit a directory, it will try to return the content
    of the file named `index.html` in that directory.

    Support request range and cache (304 status code).
    """

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if_none_match: str = ""
        if_modified_since: str = ""
        for k, v in scope["headers"]:
            if k == b"if-none-match":
                if_none_match = v.decode("latin-1")
            elif k == b"if-modified-since":
                if_modified_since = v.decode("latin-1")
        filepath = self.ensure_absolute_path(scope["path"])
        stat_result, is_file = self.check_path_is_file(filepath)
        if (
            stat_result is None  # filepath is not exist
            and filepath is not None  # Just for type check
            and not filepath.endswith(".html")  # filepath is not a html file
        ):
            filepath += ".html"
            stat_result, is_file = self.check_path_is_file(filepath)

        if stat_result is not None:
            assert filepath is not None  # Just for type check
            if is_file:
                return await self.file_response(
                    filepath, stat_result, if_none_match, if_modified_since
                )(scope, receive, send)
            if stat.S_ISDIR(stat_result.st_mode):
                url = URL(scope=scope)
                url = url.replace(scheme="", path=url.path + "/")
                return await RedirectResponse(url)(scope, receive, send)

        if self.handle_404 is None:
            raise HTTPException(404)
        else:
            return await self.handle_404(scope, receive, send)

    def ensure_absolute_path(self, path: str) -> Optional[str]:
        abspath = super().ensure_absolute_path(path)
        if abspath is not None:
            if abspath.endswith("/"):
                abspath += "index.html"
        return abspath



import datetime
import os
import re
import string
import typing
from tempfile import SpooledTemporaryFile
from urllib.parse import SplitResult, parse_qsl, urlencode, urlsplit

from .concurrency import run_in_threadpool
from .typing import Environ, Final, Literal, Scope
from .utils import parse_header

__all__ = [
    "Address",
    "MediaType",
    "ContentType",
    "Cookie",
    "URL",
    "MultiMapping",
    "MutableMultiMapping",
    "QueryParams",
    "Headers",
    "MutableHeaders",
    "UploadFile",
    "FormData",
]


T = typing.TypeVar("T")  # Any type.
KT = typing.TypeVar("KT")  # Key type.
VT = typing.TypeVar("VT")  # Value type.


class Address(typing.NamedTuple):
    host: typing.Optional[str]
    port: typing.Optional[int]


class defaultdict(dict):
    def __init__(self, default_factory, *args, **kwargs) -> None:
        self.default_factory = default_factory
        super().__init__(*args, **kwargs)

    def __missing__(self, key):
        return self.default_factory(key)


class MediaType:
    __slots__ = ("main_type", "sub_type", "options")

    def __init__(self, media_type_raw_line: str) -> None:
        full_type, self.options = parse_header(media_type_raw_line)
        self.main_type, _, self.sub_type = full_type.partition("/")

    def __str__(self) -> str:
        return (
            self.main_type
            + (f"/{self.sub_type}" if self.sub_type else "")
            + "".join(f"; {k}={v}" for k, v in self.options.items())
        )

    def __repr__(self) -> str:
        return f"<{self.__class__.__qualname__}: {self}>"

    @property
    def is_all_types(self) -> bool:
        return self.main_type == "*" and self.sub_type == "*"

    def match(self, other: str) -> bool:
        if self.is_all_types:
            return True
        other_media_type = MediaType(other)
        return self.main_type == other_media_type.main_type and (
            self.sub_type in {"*", other_media_type.sub_type}
        )


class ContentType:
    __slots__ = ("type", "options")

    def __init__(self, content_type_raw_line: str) -> None:
        self.type, self.options = parse_header(content_type_raw_line)

    def __repr__(self) -> str:
        return f"<{self.__class__.__qualname__}: {self}>"

    def __str__(self) -> str:
        return self.type + "".join(f"; {k}={v}" for k, v in self.options.items())

    def __eq__(self, other: typing.Any) -> bool:
        if not isinstance(other, str):
            return NotImplemented
        return self.type == other


_cookie_legal_chars: Final[str] = (
    string.ascii_letters + string.digits + "!#$%&'*+-.^_`|~:"
)
_cookie_is_legal_key = re.compile("[%s]+" % re.escape(_cookie_legal_chars)).fullmatch
_cookie_translator: Final[typing.Dict[int, str]] = {
    **{
        n: "\\%03o" % n
        for n in set(range(256)) - set(map(ord, _cookie_legal_chars + " ()/<=>?@[]{}"))
    },
    ord('"'): '\\"',
    ord("\\"): "\\\\",
}


class Cookie:
    def __init__(
        self,
        name: str,
        value: str,
        expires: typing.Optional[datetime.datetime] = None,
        domain: typing.Optional[str] = None,
        path: typing.Optional[str] = None,
        httponly: bool = False,
        secure: bool = False,
        max_age: int = -1,
        samesite: Literal["strict", "lax", "none"] = "lax",
    ):
        self.name = name
        self.value = value
        self.expires = expires
        self.domain = domain
        self.path = path
        self.httponly = httponly
        self.secure = secure
        self.max_age = max_age
        self.samesite = samesite

    def _quote(self, value: str) -> str:
        r"""
        Quote a string for use in a cookie header.

        If the string does not need to be double-quoted, then just return the
        string.  Otherwise, surround the string in doublequotes and quote
        (with a \) special characters.
        """
        if _cookie_is_legal_key(value):
            return value
        else:
            return '"' + value.translate(_cookie_translator) + '"'

    def __str__(self) -> str:
        parts: typing.List[str] = []
        parts.append(f"{self._quote(self.name)}={self._quote(self.value)}")

        if self.expires:
            parts.append(
                "expires=" + self.expires.strftime("%a, %d %b %Y %H:%M:%S GMT")
            )

        if self.max_age > -1:
            parts.append(f"max-age={self.max_age}")

        if self.domain:
            parts.append(f"domain={self.domain}")

        if self.path:
            parts.append(f"path={self.path}")

        if self.httponly:
            parts.append("httponly")

        if self.secure or self.samesite in ("strict", "none"):
            parts.append("secure")

        parts.append(f"samesite={self.samesite}")

        return "; ".join(parts)

    def __bytes__(self) -> bytes:
        return str(self).encode("ascii")

    def __eq__(self, other: object) -> bool:
        if isinstance(other, str):
            return str(self) == other
        if isinstance(other, bytes):
            return bytes(self) == other
        if isinstance(other, Cookie):
            return str(self) == str(other)
        return NotImplemented

    def __repr__(self) -> str:
        return f"<Cookie {self.name}: {self.value}>"


class URL:
    __slots__ = ("_url", "_components")

    def __init__(
        self,
        url: str = "",
        *,
        scope: typing.Optional[Scope] = None,
        environ: typing.Optional[Environ] = None,
        **components: typing.Any,
    ) -> None:
        if components:
            assert not url, 'Cannot set both "url" and "**components".'
            url = URL("").replace(**components).components.geturl()
        elif scope is not None:
            scheme = scope.get("scheme", "http")
            server = scope.get("server", None)
            path = scope.get("root_path", "") + scope["path"]
            query_string = scope.get("query_string", b"")

            host_header = None
            for key, value in scope["headers"]:
                if key == b"host":
                    host_header = value.decode("latin-1")
                    break
            url = self._build_url(scheme, path, query_string, server, host_header)
        elif environ is not None:
            scheme = environ["wsgi.url_scheme"]
            server = (environ["SERVER_NAME"], int(environ["SERVER_PORT"]))
            path = (
                (environ.get("SCRIPT_NAME", "") + environ.get("PATH_INFO", ""))
                .encode("latin1")
                .decode("utf8")
            )
            query_string = environ.get("QUERY_STRING", "").encode("latin-1")
            host_header = environ.get("HTTP_HOST", None)
            url = self._build_url(scheme, path, query_string, server, host_header)

        self._url = url
        self._components = urlsplit(url)

    def _build_url(
        self,
        scheme: str,
        path: str,
        query_string: bytes = b"",
        server: typing.Optional[typing.Tuple[str, typing.Optional[int]]] = None,
        host_header: typing.Optional[str] = None,
    ) -> str:
        if host_header is not None:
            url = f"{scheme}://{host_header}{path}"
        elif server is None:
            url = path
        else:
            host, port = server
            default_port = {"http": 80, "https": 443, "ws": 80, "wss": 443}[scheme]
            if port == default_port or port is None:
                url = f"{scheme}://{host}{path}"
            else:
                url = f"{scheme}://{host}:{port}{path}"

        if query_string:
            url = f"{url}?{query_string.decode()}"

        return url

    @property
    def components(self) -> SplitResult:
        return self._components

    @property
    def scheme(self) -> str:
        return self.components.scheme

    @property
    def netloc(self) -> str:
        return self.components.netloc

    @property
    def path(self) -> str:
        return self.components.path

    @property
    def query(self) -> str:
        return self.components.query

    @property
    def fragment(self) -> str:
        return self.components.fragment

    @property
    def username(self) -> typing.Optional[str]:
        return self.components.username

    @property
    def password(self) -> typing.Optional[str]:
        return self.components.password

    @property
    def hostname(self) -> typing.Optional[str]:
        return self.components.hostname

    @property
    def port(self) -> typing.Optional[int]:
        return self.components.port

    def replace(self, **kwargs: typing.Any) -> "URL":
        if (
            "username" in kwargs
            or "password" in kwargs
            or "hostname" in kwargs
            or "port" in kwargs
        ):
            hostname = kwargs.pop("hostname", None)
            port = kwargs.pop("port", self.port)
            username = kwargs.pop("username", self.username)
            password = kwargs.pop("password", self.password)

            if hostname is None:
                netloc = self.netloc
                _, _, hostname = netloc.rpartition("@")

                if hostname[-1] != "]":
                    hostname = hostname.rsplit(":", 1)[0]

            netloc = hostname
            if port is not None:
                netloc += f":{port}"
            if username is not None:
                userpass = username
                if password is not None:
                    userpass += f":{password}"
                netloc = f"{userpass}@{netloc}"

            kwargs["netloc"] = netloc

        components = self.components._replace(**kwargs)
        return self.__class__(components.geturl())

    def include_query_params(self, **kwargs: typing.Any) -> "URL":
        params: MutableMultiMapping[str, str] = MutableMultiMapping(
            parse_qsl(self.query, keep_blank_values=True)
        )
        params.update({key: str(value) for key, value in kwargs.items()})
        query = urlencode(params.multi_items())
        return self.replace(query=query)

    def replace_query_params(self, **kwargs: typing.Any) -> "URL":
        query = urlencode([(key, str(value)) for key, value in kwargs.items()])
        return self.replace(query=query)

    def remove_query_params(self, *keys: str) -> "URL":
        params: MutableMultiMapping[str, str] = MutableMultiMapping(
            parse_qsl(self.query, keep_blank_values=True)
        )
        [params.pop(key, None) for key in keys]
        query = urlencode(params.multi_items())
        return self.replace(query=query)

    def __eq__(self, other: typing.Any) -> bool:
        return str(self) == str(other)

    def __str__(self) -> str:
        return self._url

    def __repr__(self) -> str:
        url = str(self)
        if self.password:
            url = str(self.replace(password="********"))
        return f"{self.__class__.__name__}({repr(url)})"


class MultiMapping(typing.Generic[KT, VT], typing.Mapping[KT, VT]):
    __slots__ = ("_dict", "_list")

    def __init__(
        self,
        raw: typing.Optional[
            typing.Union[
                typing.Mapping[KT, VT],
                typing.Iterable[typing.Tuple[KT, VT]],
            ]
        ] = None,
    ) -> None:
        _items: typing.List[typing.Tuple[KT, VT]]
        if raw is None:
            _items = []
        elif isinstance(raw, MultiMapping):
            _items = typing.cast(
                typing.List[typing.Tuple[KT, VT]], list(raw.multi_items())
            )
        elif isinstance(raw, typing.Mapping):
            _items = typing.cast(typing.List[typing.Tuple[KT, VT]], list(raw.items()))
        else:
            _items = list(raw)

        self._dict = dict(_items)
        self._list = _items

    def __getitem__(self, key: KT) -> VT:
        return self._dict[key]

    def __iter__(self) -> typing.Iterator[KT]:
        return iter(self._dict)

    def __len__(self) -> int:
        return len(self._dict)

    def getlist(self, key: KT) -> typing.List[VT]:
        return [item_value for item_key, item_value in self._list if item_key == key]

    def multi_items(self) -> typing.List[typing.Tuple[KT, VT]]:
        return list(self._list)

    def __eq__(self, other: typing.Any) -> bool:
        return isinstance(other, type(self)) and sorted(self._list) == sorted(
            other._list
        )

    def __repr__(self) -> str:
        class_name = self.__class__.__name__
        items = self.multi_items()
        return f"{class_name}({items!r})"


class MutableMultiMapping(
    typing.Generic[KT, VT], MultiMapping[KT, VT], typing.MutableMapping[KT, VT]
):
    __slots__ = MultiMapping.__slots__

    def __setitem__(self, key: KT, value: VT) -> None:
        indexes = tuple(index for index, kv in enumerate(self._list) if kv[0] == key)
        if indexes:
            frist_index = indexes[0]
            for index in reversed(indexes):
                if index == frist_index:
                    self._list[index] = (key, value)
                else:
                    del self._list[index]
        else:
            self._list.append((key, value))
        self._dict[key] = value

    def __delitem__(self, key: KT) -> None:
        _list = self._list[:]
        self._list.clear()
        self._list.extend((k, v) for k, v in _list if k != key)
        del self._dict[key]

    def setlist(self, key: KT, values: typing.Sequence[VT]) -> None:
        if values:
            self._list = [
                *((k, v) for k, v in self._list if k != key),
                *((key, value) for value in values),
            ]
            self._dict[key] = values[-1]
        elif key in self:
            del self[key]

    def poplist(self, key: KT) -> typing.List[VT]:
        values = [v for k, v in self._list if k == key]
        try:
            del self[key]
        except KeyError:
            pass
        return values

    def append(self, key: KT, value: VT) -> None:
        self._list.append((key, value))
        self._dict[key] = value


class QueryParams(MultiMapping[str, str]):
    """
    An immutable MutableMultiMapping.
    """

    __slots__ = ("_dict", "_list")

    def __init__(
        self,
        raw: typing.Optional[
            typing.Union[
                "MultiMapping[str, str]",
                typing.Mapping[str, str],
                typing.Iterable[typing.Tuple[str, str]],
                str,
                bytes,
            ]
        ] = None,
    ) -> None:
        if isinstance(raw, str):
            super().__init__(parse_qsl(raw, keep_blank_values=True))
        elif isinstance(raw, bytes):
            super().__init__(parse_qsl(raw.decode("latin-1"), keep_blank_values=True))
        else:
            super().__init__(raw)

    def __str__(self) -> str:
        return urlencode(self._list)

    def __repr__(self) -> str:
        class_name = self.__class__.__name__
        query_string = str(self)
        return f"{class_name}({query_string!r})"


class Headers(typing.Mapping[str, str]):
    __slots__ = ("_dict",)

    def __init__(
        self,
        headers: typing.Optional[
            typing.Union[
                typing.Mapping[str, str],
                typing.Iterable[typing.Tuple[str, str]],
            ]
        ] = None,
    ) -> None:
        store: typing.Dict[str, str] = {}
        items: typing.Iterable[typing.Tuple[str, str]]
        if isinstance(headers, typing.Mapping):
            items = typing.cast(
                typing.Iterable[typing.Tuple[str, str]], headers.items()
            )
        elif headers is None:
            items = ()
        else:
            items = headers
        for key, value in items:
            key = key.lower()
            if key in store:
                store[key] = f"{store[key]}, {value}"
            else:
                store[key] = value

        self._dict = store

    def __getitem__(self, key: str) -> str:
        return self._dict[key.lower()]

    def __iter__(self) -> typing.Iterator[str]:
        return self._dict.__iter__()

    def __len__(self) -> int:
        return self._dict.__len__()


class MutableHeaders(Headers, typing.MutableMapping[str, str]):
    __slots__ = Headers.__slots__

    def __setitem__(self, key: str, value: str) -> None:
        if "\n" in key or "\r" in key or "\0" in key:
            raise ValueError("Header names must not contain control characters.")
        if "\n" in value or "\r" in value or "\0" in value:
            raise ValueError("Header values must not contain control characters.")
        self._dict[key.lower()] = value

    def __delitem__(self, key: str) -> None:
        del self._dict[key.lower()]

    def append(self, key: str, value: str) -> None:
        if key in self:
            self[key] = f"{self[key]}, {value}"
        else:
            self[key] = value


class UploadFile:
    """
    An uploaded file included as part of the request data.
    """

    __slots__ = ("filename", "headers", "content_type", "file")

    spool_max_size = 1024 * 1024

    def __init__(self, filename: str, headers: Headers) -> None:
        self.filename = filename
        self.headers = headers
        self.content_type = headers.get("content-type", "")
        self.file = SpooledTemporaryFile(max_size=self.spool_max_size, mode="w+b")

    @property
    def in_memory(self) -> bool:
        rolled_to_disk = getattr(self.file, "_rolled", True)
        return not rolled_to_disk

    def write(self, data: bytes) -> None:
        self.file.write(data)

    async def awrite(self, data: bytes) -> None:
        if self.in_memory:
            self.write(data)
        else:
            await run_in_threadpool(self.write, data)

    def read(self, size: int = -1) -> bytes:
        return self.file.read(size)

    async def aread(self, size: int = -1) -> bytes:
        if self.in_memory:
            return self.read(size)
        return await run_in_threadpool(self.read, size)

    def seek(self, offset: int) -> None:
        self.file.seek(offset)

    async def aseek(self, offset: int) -> None:
        if self.in_memory:
            self.seek(offset)
        else:
            await run_in_threadpool(self.seek, offset)

    def close(self) -> None:
        self.file.close()

    async def aclose(self) -> None:
        if self.in_memory:
            self.close()
        else:
            await run_in_threadpool(self.close)

    def save(self, filepath: str) -> None:
        """
        Save file to disk.
        """
        # from shutil.COPY_BUFSIZE
        copy_bufsize = 1024 * 1024 if os.name == "nt" else 64 * 1024
        file_position = self.file.tell()
        self.file.seek(0, 0)
        try:
            with open(filepath, "wb+") as target_file:
                source_read = self.file.read
                target_write = target_file.write
                while True:
                    buf = source_read(copy_bufsize)
                    if not buf:
                        break
                    target_write(buf)
        finally:
            self.file.seek(file_position)

    async def asave(self, filepath: str) -> None:
        """
        Save file to disk, work in threading pool.
        """
        await run_in_threadpool(self.save, filepath)


class FormData(MultiMapping[str, typing.Union[str, UploadFile]]):
    """
    An immutable MultiMapping, containing both file uploads and text input.
    """

    __slots__ = MultiMapping.__slots__

    def close(self) -> None:
        for key, value in self.multi_items():
            if isinstance(value, UploadFile):
                value.close()

    async def aclose(self) -> None:
        for key, value in self.multi_items():
            if isinstance(value, UploadFile):
                await value.aclose()



from typing import AsyncIterable, Iterable, List, Optional, Tuple, Type, TypeVar, Union

from .datastructures import Headers
from .exceptions import RequestEntityTooLarge
from .multipart import (
    Data,
    Epilogue,
    Field,
    File,
    MultipartDecoder,
    NeedData,
    safe_decode,
)
from .typing import Protocol, runtime_checkable


@runtime_checkable
class SyncUploadFileInterface(Protocol):
    def __init__(self, filename: str, headers: Headers) -> None: ...

    def write(self, data: bytes) -> None: ...

    def seek(self, offset: int) -> None: ...


@runtime_checkable
class AsyncUploadFileInterface(Protocol):
    def __init__(self, filename: str, headers: Headers) -> None: ...

    async def awrite(self, data: bytes) -> None: ...

    async def aseek(self, offset: int) -> None: ...


_SyncUploadFile = TypeVar("_SyncUploadFile", bound=SyncUploadFileInterface)
_AsyncUploadFile = TypeVar("_AsyncUploadFile", bound=AsyncUploadFileInterface)


async def parse_async_stream(
    stream: AsyncIterable[bytes],
    boundary: bytes,
    charset: str,
    *,
    file_factory: Type[_AsyncUploadFile],
    max_form_parts: int = 324,
    max_form_memory_size: Union[int, None] = None,
) -> List[Tuple[str, Union[str, _AsyncUploadFile]]]:
    """
    Parse an asynchronous stream in multipart format

    ```python
    for field_name, field_or_file in await parse_async_stream(stream, boundary, charset):
        print(field_name, field_or_file)
    ```
    """
    parser = MultipartDecoder(boundary, charset)
    field_name = ""
    data = bytearray()
    file: Optional[_AsyncUploadFile] = None
    form_parts_count = 0
    form_memory_size_count = 0

    items: List[Tuple[str, Union[str, _AsyncUploadFile]]] = []

    async for chunk in stream:
        parser.receive_data(chunk)
        while True:
            event = parser.next_event()
            if isinstance(event, (Epilogue, NeedData)):
                break
            elif isinstance(event, Field):
                field_name = event.name
            elif isinstance(event, File):
                field_name = event.name
                file = file_factory(event.filename, event.headers)
            elif isinstance(event, Data):
                if file is None:
                    data.extend(event.data)

                    # Check if we have exceeded the maximum memory size
                    form_memory_size_count += len(event.data)
                    if (
                        max_form_memory_size is not None
                        and form_memory_size_count > max_form_memory_size
                    ):
                        raise RequestEntityTooLarge()
                else:
                    await file.awrite(event.data)

                if not event.more_data:
                    if file is None:
                        items.append((field_name, safe_decode(data, charset)))
                        data.clear()
                    else:
                        await file.aseek(0)
                        items.append((field_name, file))
                        file = None

                    # Check if we have exceeded the maximum number of form parts
                    form_parts_count += 1
                    if form_parts_count > max_form_parts:
                        raise RequestEntityTooLarge()
    return items


def parse_stream(
    stream: Iterable[bytes],
    boundary: bytes,
    charset: str,
    *,
    file_factory: Type[_SyncUploadFile],
    max_form_parts: int = 324,
    max_form_memory_size: Union[int, None] = None,
) -> List[Tuple[str, Union[str, _SyncUploadFile]]]:
    """
    Parse a synchronous stream in multipart format

    ```python
    for field_name, field_or_file in parse_stream(stream, boundary, charset):
        print(field_name, field_or_file)
    ```
    """
    parser = MultipartDecoder(boundary, charset)
    field_name = ""
    data = bytearray()
    file: Optional[_SyncUploadFile] = None
    form_parts_count = 0
    form_memory_size_count = 0

    items: List[Tuple[str, Union[str, _SyncUploadFile]]] = []

    for chunk in stream:
        parser.receive_data(chunk)
        while True:
            event = parser.next_event()
            if isinstance(event, (Epilogue, NeedData)):
                break
            elif isinstance(event, Field):
                field_name = event.name
            elif isinstance(event, File):
                field_name = event.name
                file = file_factory(event.filename, event.headers)
            elif isinstance(event, Data):
                if file is None:
                    data.extend(event.data)

                    # Check if we have exceeded the maximum memory size
                    form_memory_size_count += len(event.data)
                    if (
                        max_form_memory_size is not None
                        and form_memory_size_count > max_form_memory_size
                    ):
                        raise RequestEntityTooLarge()
                else:
                    file.write(event.data)

                if not event.more_data:
                    if file is None:
                        items.append((field_name, safe_decode(data, charset)))
                        data.clear()
                    else:
                        file.seek(0)
                        items.append((field_name, file))
                        file = None

                    # Check if we have exceeded the maximum number of form parts
                    form_parts_count += 1
                    if form_parts_count > max_form_parts:
                        raise RequestEntityTooLarge()
    return items



import json
from typing import Any, Dict, Iterator, Mapping, Optional
from urllib.parse import parse_qsl

from baize.datastructures import (
    URL,
    Address,
    FormData,
    Headers,
    QueryParams,
    UploadFile,
)
from baize.exceptions import MalformedJSON, MalformedMultipart, UnsupportedMediaType
from baize.multipart_helper import parse_stream as parse_multipart
from baize.requests import MoreInfoFromHeaderMixin
from baize.typing import Environ, StartResponse
from baize.utils import cached_property


class HTTPConnection(Mapping[str, Any], MoreInfoFromHeaderMixin):
    """
    A base class for incoming HTTP connections.

    It is a valid Mapping type that allows you to directly
    access the values in any WSGI `environ` dictionary.
    """

    def __init__(
        self, environ: Environ, start_response: Optional[StartResponse] = None
    ) -> None:
        self._environ = environ
        self._start_response = start_response

    def __getitem__(self, key: str) -> Any:
        return self._environ[key]

    def __iter__(self) -> Iterator[str]:
        return iter(self._environ)

    def __len__(self) -> int:
        return len(self._environ)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self._environ == other._environ
            and self._start_response == other._start_response
        )

    @cached_property
    def client(self) -> Address:
        """
        Client's IP and Port.

        Note that this depends on the `REMOTE_ADDR` and `REMOTE_PORT` values
        given by the WSGI Server, and is not necessarily accurate.
        """
        if self.get("REMOTE_ADDR") and self.get("REMOTE_PORT"):
            return Address(self["REMOTE_ADDR"], int(self["REMOTE_PORT"]))
        return Address(host=None, port=None)

    @cached_property
    def url(self) -> URL:
        """
        The full URL of this request.
        """
        return URL(environ=self._environ)

    @cached_property
    def path_params(self) -> Dict[str, Any]:
        """
        The path parameters parsed by the framework.
        """
        return self.get("PATH_PARAMS", {})

    @cached_property
    def query_params(self) -> QueryParams:
        """
        Query parameter. It is a multi-value mapping.
        """
        return QueryParams(self["QUERY_STRING"])

    @cached_property
    def headers(self) -> Headers:
        """
        A read-only case-independent mapping.

        Note that in its internal storage, all keys are in lower case.
        """
        return Headers(
            (
                (key[5:] if key.startswith("HTTP_") else key).lower().replace("_", "-"),
                value,
            )
            for key, value in self._environ.items()
            if key.startswith("HTTP_") or key in ("CONTENT_TYPE", "CONTENT_LENGTH")
        )


class Request(HTTPConnection):
    def __init__(
        self, environ: Environ, start_response: Optional[StartResponse] = None
    ) -> None:
        super().__init__(environ, start_response)
        self._stream_consumed = False

    @property
    def method(self) -> str:
        """
        HTTP method. Uppercase string.
        """
        return self["REQUEST_METHOD"]

    def stream(self, chunk_size: int = 4096 * 16) -> Iterator[bytes]:
        """
        Streaming read request body. e.g. `for chunk in request.stream(): ...`

        If you access `.stream()` then the byte chunks are provided
        without storing the entire body to memory. Any subsequent
        calls to `.body`, `.form`, or `.json` will raise an error.
        """
        if "body" in self.__dict__:
            yield self.body
            return

        if self._stream_consumed:
            raise RuntimeError("Stream consumed")

        self._stream_consumed = True
        body = self._environ["wsgi.input"]
        while True:
            chunk = body.read(chunk_size)
            if not chunk:
                return
            yield chunk

    @cached_property
    def body(self) -> bytes:
        """
        Read all the contents of the request body into the memory and return it.
        """
        return b"".join([chunk for chunk in self.stream()])

    @cached_property
    def json(self) -> Any:
        """
        Call `self.body` and use `json.loads` parse it.

        If `content_type` is not equal to `application/json`,
        an HTTPExcption exception will be thrown.
        """
        if self.content_type == "application/json":
            try:
                return json.loads(
                    self.body.decode(self.content_type.options.get("charset", "utf8"))
                )
            except json.JSONDecodeError as exc:
                raise MalformedJSON(str(exc)) from None

        raise UnsupportedMediaType("application/json")

    def _parse_multipart(self, boundary: bytes, charset: str) -> FormData:
        return FormData(
            parse_multipart(self.stream(), boundary, charset, file_factory=UploadFile)
        )

    @cached_property
    def form(self) -> FormData:
        """
        Parse the data in the form format and return it as a multi-value mapping.

        If `content_type` is equal to `multipart/form-data`, it will directly
        perform streaming analysis, and subsequent calls to `self.body`
        or `self.json` will raise errors.

        If `content_type` is not equal to `multipart/form-data` or
        `application/x-www-form-urlencoded`, an HTTPExcption exception will be thrown.
        """
        if self.content_type == "multipart/form-data":
            charset = self.content_type.options.get("charset", "utf8")
            if "boundary" not in self.content_type.options:
                raise MalformedMultipart("Missing boundary in header content-type")
            boundary = self.content_type.options["boundary"].encode("latin-1")
            return self._parse_multipart(boundary, charset)
        if self.content_type == "application/x-www-form-urlencoded":
            body = self.body.decode(
                encoding=self.content_type.options.get("charset", "latin-1")
            )
            return FormData(parse_qsl(body, keep_blank_values=True))

        raise UnsupportedMediaType(
            "multipart/form-data, application/x-www-form-urlencoded"
        )

    def close(self) -> None:
        """
        Close all temporary files in the `self.form`.

        This can always be called, regardless of whether you use form or not.
        """
        if "form" in self.__dict__:
            self.form.close()



from datetime import datetime, timezone
from email.utils import parsedate_to_datetime
from http import cookies as http_cookies
from typing import Dict, List, Optional

try:
    from mypy_extensions import mypyc_attr, trait
except ImportError:  # pragma: no cover

    def trait(cls):  # type: ignore
        return cls

    def mypyc_attr(*attrs, **kwattrs):  # type: ignore
        return lambda x: x


from .datastructures import URL, ContentType, Headers, MediaType
from .utils import cached_property


@mypyc_attr(allow_interpreted_subclasses=True)
class MoreInfoFromHeaderMixin:
    """
    Parse more information from the header for quick use
    """

    @cached_property
    def headers(self) -> Headers:
        raise NotImplementedError

    @cached_property
    def accepted_types(self) -> List[MediaType]:
        """
        Request's accepted types
        """
        return [
            MediaType(token)
            for token in self.headers.get("Accept", "*/*").split(",")
            if token.strip()
        ]

    def accepts(self, media_type: str) -> bool:
        """
        e.g. `request.accepts("application/json")`
        """
        return any(
            accepted_type.match(media_type) for accepted_type in self.accepted_types
        )

    @cached_property
    def content_type(self) -> ContentType:
        """
        Request's content-type
        """
        return ContentType(self.headers.get("content-type", ""))

    @cached_property
    def content_length(self) -> Optional[int]:
        """
        Request's content-length
        """
        if self.headers.get("transfer-encoding", "") == "chunked":
            return None

        content_length = self.headers.get("content-length", None)
        if content_length is None:
            return None

        try:
            return max(0, int(content_length))
        except (ValueError, TypeError):
            return None

    @cached_property
    def cookies(self) -> Dict[str, str]:
        """
        Returns cookies in as a `dict`.

        NOTE: Modifications to this dictionary will not affect the
        response value. In fact, this value should not be modified.
        """
        cookies: Dict[str, str] = {}
        cookie_header = self.headers.get("cookie", "")

        # This function has been adapted from Django 3.1.0.
        # Note: we are explicitly _NOT_ using `SimpleCookie.load` because it is based
        # on an outdated spec and will fail on lots of input we want to support
        for chunk in cookie_header.split(";"):
            if not chunk:
                continue
            if "=" in chunk:
                key, val = chunk.split("=", 1)
            else:
                # Assume an empty name per
                # https://bugzilla.mozilla.org/show_bug.cgi?id=169091
                key, val = "", chunk
            key, val = key.strip(), val.strip()
            if key or val:
                # unquote using Python's algorithm.
                cookies[key] = http_cookies._unquote(val)  # type: ignore
        return cookies

    @cached_property
    def date(self) -> Optional[datetime]:
        """
        The sending time of the request.

        NOTE: The datetime object is timezone-aware.
        """
        value = self.headers.get("date", None)
        if value is None:
            return None

        try:
            date = parsedate_to_datetime(value)
        except (TypeError, ValueError):
            return None

        if date.tzinfo is None:
            return date.replace(tzinfo=timezone.utc)

        return date

    @cached_property
    def referrer(self) -> Optional[URL]:
        """
        The `Referer` HTTP request header contains an absolute or partial address
        of the page making the request.
        """
        referrer = self.headers.get("referer", None)
        if referrer is None:
            return None

        return URL(url=referrer)



import asyncio
import functools
from typing import (
    Any,
    AsyncGenerator,
    AsyncIterator,
    Awaitable,
    Callable,
    MutableMapping,
)

from ..datastructures import Headers
from ..typing import ASGIApp, Scope, Receive, Send, Message
from .requests import Request
from .responses import Response, StreamingResponse


class PipeStream(AsyncIterator[bytes]):
    def __init__(self) -> None:
        self._queue: asyncio.Queue[bytes] = asyncio.Queue(maxsize=1)
        self._pushed_eof = False

    async def push(self, chunk: bytes) -> None:
        if self._pushed_eof:
            raise RuntimeError("Cannot push chunk after push EOF.")  # pragma: no cover
        await self._queue.put(chunk)

    async def push_eof(self) -> None:
        self._pushed_eof = True

    async def __anext__(self) -> bytes:
        if self._pushed_eof and self._queue.empty():
            raise StopAsyncIteration
        return await self._queue.get()


class NextRequest(Request, MutableMapping[str, Any]):
    def __setitem__(self, name: str, value: Any) -> None:
        self._scope[name] = value

    def __delitem__(self, name: str) -> None:
        del self._scope[name]

    def stream(self) -> AsyncIterator[bytes]:
        raise RuntimeError("Cannot read request body in middleware.")


class NextResponse(StreamingResponse):
    """
    This is a response object for middleware.
    """

    async def render_stream(self) -> AsyncGenerator[bytes, None]:
        async for chunk in self.iterable:
            yield chunk

    @classmethod
    async def from_app(cls, app: ASGIApp, request: NextRequest) -> "NextResponse":
        """
        This is a helper method to convert a ASGI application into a NextResponse object.
        """
        status_code = 200
        headers = Headers()
        body = PipeStream()

        async def send(message: Message) -> None:
            nonlocal status_code
            nonlocal headers
            if message["type"] == "http.response.start":
                status_code = message["status"]
                headers = Headers(
                    [
                        (k.decode("latin-1"), v.decode("latin-1"))
                        for k, v in message.get("headers", [])
                    ]
                )
            elif message["type"] == "http.response.body":
                await body.push(message.get("body", b""))
                if not message.get("more_body", False):
                    await body.push_eof()

        await app(request, request._receive, send)
        return NextResponse(body, status_code, headers)


def middleware(
    handler: Callable[
        [NextRequest, Callable[[NextRequest], Awaitable[NextResponse]]],
        Awaitable[Response],
    ],
) -> Callable[[ASGIApp], ASGIApp]:
    """
    This can turn a callable object into a middleware for ASGI application.

    ```python
    @middleware
    async def m(
        request: NextRequest, next_call: Callable[[NextRequest], Awaitable[NextResponse]]
    ) -> Response:
        ...
        response = await next_call(request)
        ...
        return response

    @m
    @request_response
    async def v(request: Request) -> Response:
        ...

    # OR

    @m
    async def asgi(scope: Scope, receive: Receive, send: Send) -> None:
        ...
    ```
    """

    @functools.wraps(handler)
    def d(app: ASGIApp) -> ASGIApp:
        """
        This is the actual middleware.
        """

        @functools.wraps(app)
        async def asgi(scope: Scope, receive: Receive, send: Send) -> None:
            request = NextRequest(scope, receive, send)

            async def next_call(request: NextRequest) -> NextResponse:
                return await NextResponse.from_app(app, request)

            response = await handler(request, next_call)
            await response(scope, receive, send)

        return asgi

    return d



import abc
import asyncio
import json
import os
import stat
from mimetypes import guess_type
from random import choices as random_choices
from typing import (
    Any,
    AsyncGenerator,
    AsyncIterable,
    Dict,
    Generic,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
)

from baize.concurrency import run_in_threadpool
from baize.datastructures import URL
from baize.exceptions import MalformedRangeHeader, RangeNotSatisfiable
from baize.responses import (
    BaseResponse,
    FileResponseMixin,
    build_bytes_from_sse,
    iri_to_uri,
)
from baize.typing import Protocol, Receive, Scope, Send, ServerSentEvent

from .helper import send_http_body, send_http_start


class Response(BaseResponse):
    """
    The parent class of all responses, whose objects can be used directly as ASGI
    application.
    """

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        self.headers["content-length"] = "0"
        await send_http_start(send, self.status_code, self.list_headers(as_bytes=True))
        return await send_http_body(send)


_ContentType = TypeVar("_ContentType")


class SmallResponse(Response, abc.ABC, Generic[_ContentType]):
    """
    Abstract base class for small response objects.
    """

    media_type: str = ""
    charset = "utf-8"

    def __init__(
        self,
        content: _ContentType,
        status_code: int = 200,
        headers: Optional[Mapping[str, str]] = None,
        media_type: Optional[str] = None,
        charset: Optional[str] = None,
    ) -> None:
        super().__init__(status_code, headers)
        self.content = content
        self.media_type = media_type or self.media_type
        self.charset = charset or self.charset

    @abc.abstractmethod
    async def render(self, content: _ContentType) -> bytes:
        raise NotImplementedError

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        body = await self.render(self.content)
        if body and "content-length" not in self.headers:
            content_length = str(len(body))
            self.headers["content-length"] = content_length
        content_type = self.media_type
        if content_type and "content-type" not in self.headers:
            if content_type.startswith("text/"):
                content_type += "; charset=" + self.charset
            self.headers["content-type"] = content_type
        await send_http_start(send, self.status_code, self.list_headers(as_bytes=True))
        await send_http_body(send, body)


class PlainTextResponse(SmallResponse[Union[bytes, str]]):
    media_type = "text/plain"

    async def render(self, content: Union[bytes, str]) -> bytes:
        return content.encode(self.charset) if isinstance(content, str) else content


class HTMLResponse(PlainTextResponse):
    media_type = "text/html"


class JSONResponse(SmallResponse[Any]):
    """
    `**kwargs` is used to accept all the parameters that `json.loads` can accept.
    """

    media_type = "application/json"

    def __init__(
        self,
        content: Any,
        status_code: int = 200,
        headers: Optional[Mapping[str, str]] = None,
        **kwargs: Any,
    ) -> None:
        self.json_kwargs: Dict[str, Any] = {
            "ensure_ascii": False,
            "allow_nan": False,
            "indent": None,
            "separators": (",", ":"),
            "default": None,
        }
        self.json_kwargs.update(**kwargs)
        super().__init__(content, status_code=status_code, headers=headers)

    async def render(self, content: Any) -> bytes:
        return json.dumps(content, **self.json_kwargs).encode(self.charset)


class RedirectResponse(Response):
    def __init__(
        self,
        url: Union[str, URL],
        status_code: int = 307,
        headers: Optional[Mapping[str, str]] = None,
    ) -> None:
        super().__init__(status_code=status_code, headers=headers)
        self.headers["location"] = iri_to_uri(str(url))


class StreamingResponse(Response, abc.ABC, Generic[_ContentType]):
    def __init__(
        self,
        iterable: AsyncIterable[_ContentType],
        status_code: int = 200,
        headers: Optional[Mapping[str, str]] = None,
    ) -> None:
        super().__init__(status_code, headers)
        self.iterable = iterable
        self._client_closed = False

    @abc.abstractmethod
    async def render_stream(self) -> AsyncGenerator[bytes, None]:
        raise NotImplementedError
        yield

    async def wait_close(self, receive: Receive) -> None:
        while not self._client_closed:
            message = await receive()
            self._client_closed = message["type"] == "http.disconnect"

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        await send_http_start(send, self.status_code, self.list_headers(as_bytes=True))
        wait_close_future = asyncio.ensure_future(self.wait_close(receive))
        generator = self.render_stream()
        try:
            while not self._client_closed:
                chunk = await generator.asend(None)
                await send_http_body(send, chunk, more_body=True)
        except StopAsyncIteration:
            pass
        finally:
            wait_close_future.cancel()
            await generator.aclose()

        return await send_http_body(send)


class StreamResponse(StreamingResponse[bytes]):
    def __init__(
        self,
        iterable: AsyncIterable[bytes],
        status_code: int = 200,
        headers: Optional[Mapping[str, str]] = None,
        content_type: str = "application/octet-stream",
    ) -> None:
        super().__init__(iterable, status_code, headers)
        self.headers["Content-Type"] = content_type

    async def render_stream(self) -> AsyncGenerator[bytes, None]:
        try:
            async for chunk in self.iterable:
                yield chunk
        finally:
            g = self.iterable
            if hasattr(g, "aclose"):
                await g.aclose()  # type: ignore


class SendEventResponse(StreamingResponse[ServerSentEvent]):
    """
    Server-sent events response.

    When the cilent closes the connection, the generator will be closed.
    Use `try-finally` to clean up resources.

    ```python
    async def generator():
        try:
            while True:
                yield ServerSentEvent()
        finally:
            print("generator closed")

    response = SendEventResponse(generator())
    ```
    """

    required_headers = {
        "Cache-Control": "no-cache",
        "Connection": "keep-alive",
        "Content-Type": "text/event-stream",
    }

    def __init__(
        self,
        iterable: AsyncIterable[ServerSentEvent],
        status_code: int = 200,
        headers: Optional[Mapping[str, str]] = None,
        *,
        ping_interval: float = 3,
        charset: str = "utf-8",
    ) -> None:
        if headers:
            headers = {**self.required_headers, **headers}
        else:
            headers = dict(self.required_headers)
        headers["Content-Type"] += f"; charset={charset}"
        super().__init__(iterable, status_code, headers)
        self.ping_interval = ping_interval
        self.charset = charset

    async def render_stream(self) -> AsyncGenerator[bytes, None]:
        q: "asyncio.Queue[ServerSentEvent | None]" = asyncio.Queue(maxsize=1)

        should_stop = False

        async def push() -> None:
            nonlocal should_stop

            try:
                i = self.iterable.__aiter__()

                while not should_stop:
                    try:
                        await q.put(await i.__anext__())
                    except StopAsyncIteration:
                        should_stop = True
            finally:
                await q.put(None)
                g = self.iterable
                if hasattr(g, "aclose"):
                    await g.aclose()  # type: ignore

        push_future = asyncio.ensure_future(push())

        try:
            while not (push_future.done() and q.empty()):
                try:
                    event = await asyncio.wait_for(q.get(), timeout=self.ping_interval)
                    if event is None:
                        break
                    yield build_bytes_from_sse(event, self.charset)
                except asyncio.TimeoutError:
                    yield b": ping\n\n"
        finally:
            should_stop = True
            while not q.empty():
                q.get_nowait()  # pragma: no cover
            if not push_future.cancel():
                exc = push_future.exception()
                if exc is not None:
                    raise exc


class Sendfile(Protocol):
    async def __call__(
        self,
        file_descriptor: int,
        offset: Optional[int] = None,
        count: Optional[int] = None,
        more_body: bool = False,
    ) -> None:
        pass


if os.name == "nt":  # pragma: py-no-win32

    async def open_for_sendfile(
        path: Union[str, bytes, "os.PathLike[str]", "os.PathLike[bytes]"],
    ) -> int:
        return await run_in_threadpool(os.open, path, os.O_RDONLY | os.O_BINARY)  # type: ignore[attr-defined]

else:  # pragma: py-win32

    async def open_for_sendfile(
        path: Union[str, bytes, "os.PathLike[str]", "os.PathLike[bytes]"],
    ) -> int:
        return await run_in_threadpool(os.open, path, os.O_RDONLY)


class FileResponse(Response, FileResponseMixin):
    """
    File response.

    It will automatically determine whether to send only headers
    and the range of files that need to be sent.
    """

    def __init__(
        self,
        filepath: str,
        headers: Optional[Mapping[str, str]] = None,
        content_type: Optional[str] = None,
        download_name: Optional[str] = None,
        stat_result: Optional[os.stat_result] = None,
        chunk_size: int = 4096 * 64,
    ) -> None:
        super().__init__(headers=headers)
        self.filepath = filepath
        self.content_type = (
            content_type
            or guess_type(download_name or os.path.basename(filepath))[0]
            or "application/octet-stream"
        )
        self.download_name = download_name
        self.stat_result = stat_result or os.stat(filepath)
        if stat.S_ISDIR(self.stat_result.st_mode):
            raise IsADirectoryError(f"{filepath} is a directory")
        self.chunk_size = chunk_size
        self.headers.update(
            self.generate_common_headers(
                self.filepath, self.content_type, self.download_name, self.stat_result
            )
        )

    def create_send_or_zerocopy(self, scope: Scope, send: Send) -> Sendfile:
        """
        https://asgi.readthedocs.io/en/latest/extensions.html#zero-copy-send
        """
        if (
            "extensions" in scope
            and "http.response.zerocopysend" in scope["extensions"]
        ):  # pragma: no cover

            async def sendfile(
                file_descriptor: int,
                offset: Optional[int] = None,
                count: Optional[int] = None,
                more_body: bool = False,
            ) -> None:
                message = {
                    "type": "http.response.zerocopysend",
                    "file": file_descriptor,
                    "more_body": more_body,
                }
                if offset is not None:
                    message["offset"] = offset
                if count is not None:
                    message["count"] = count
                await send(message)

            return sendfile
        else:

            async def fake_sendfile(
                file_descriptor: int,
                offset: Optional[int] = None,
                count: Optional[int] = None,
                more_body: bool = False,
            ) -> None:
                if offset is not None:
                    await run_in_threadpool(
                        os.lseek, file_descriptor, offset, os.SEEK_SET
                    )

                here = 0
                should_stop = False
                if count is None:
                    length = self.chunk_size
                    while not should_stop:
                        data = await run_in_threadpool(os.read, file_descriptor, length)
                        if len(data) == length:
                            await send_http_body(send, data, more_body=True)
                        else:
                            await send_http_body(send, data, more_body=more_body)
                            should_stop = True
                else:
                    while not should_stop:
                        length = min(self.chunk_size, count - here)
                        should_stop = length == count - here
                        here += length
                        data = await run_in_threadpool(os.read, file_descriptor, length)
                        await send_http_body(
                            send, data, more_body=more_body if should_stop else True
                        )

            return fake_sendfile

    async def handle_all(
        self, send_header_only: bool, file_size: int, scope: Scope, send: Send
    ) -> None:
        self.headers["content-type"] = str(self.content_type)
        self.headers["content-length"] = str(file_size)
        await send_http_start(send, 200, self.list_headers(as_bytes=True))
        if send_header_only:
            return await send_http_body(send)

        sendfile = self.create_send_or_zerocopy(scope, send)
        file_descriptor = await open_for_sendfile(self.filepath)
        try:
            await sendfile(file_descriptor)
        finally:
            await run_in_threadpool(os.close, file_descriptor)

    async def handle_single_range(
        self,
        send_header_only: bool,
        file_size: int,
        scope: Scope,
        send: Send,
        start: int,
        end: int,
    ) -> None:
        self.headers["content-range"] = f"bytes {start}-{end - 1}/{file_size}"
        self.headers["content-type"] = str(self.content_type)
        self.headers["content-length"] = str(end - start)
        await send_http_start(send, 206, self.list_headers(as_bytes=True))
        if send_header_only:
            return await send_http_body(send)

        sendfile = self.create_send_or_zerocopy(scope, send)
        file_descriptor = await open_for_sendfile(self.filepath)
        try:
            await sendfile(file_descriptor, start, end - start)
        finally:
            await run_in_threadpool(os.close, file_descriptor)

    async def handle_several_ranges(
        self,
        send_header_only: bool,
        file_size: int,
        scope: Scope,
        send: Send,
        ranges: Sequence[Tuple[int, int]],
    ) -> None:
        boundary = "".join(random_choices("abcdefghijklmnopqrstuvwxyz0123456789", k=13))
        self.headers["content-type"] = f"multipart/byteranges; boundary={boundary}"
        content_length, generate_headers = self.generate_multipart(
            ranges, boundary, file_size, self.content_type
        )
        self.headers["content-length"] = str(content_length)
        await send_http_start(send, 206, self.list_headers(as_bytes=True))
        if send_header_only:
            return await send_http_body(send)
        sendfile = self.create_send_or_zerocopy(scope, send)
        file_descriptor = await open_for_sendfile(self.filepath)
        try:
            for start, end in ranges:
                await send_http_body(send, generate_headers(start, end), more_body=True)
                await sendfile(file_descriptor, start, end - start, True)
                await send_http_body(send, b"\n", more_body=True)
            return await send_http_body(send, f"--{boundary}--\n".encode("ascii"))
        finally:
            await run_in_threadpool(os.close, file_descriptor)

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        send_header_only = scope["method"] == "HEAD"

        stat_result = self.stat_result
        file_size = stat_result.st_size

        http_range, http_if_range = "", ""
        for key, value in scope["headers"]:
            if key == b"range":
                http_range = value.decode("latin-1")
            elif key == b"if-range":
                http_if_range = value.decode("latin-1")

        if http_range == "" or (
            http_if_range != "" and not self.judge_if_range(http_if_range, stat_result)
        ):
            return await self.handle_all(send_header_only, file_size, scope, send)

        try:
            ranges = self.parse_range(http_range, file_size)
        except (MalformedRangeHeader, RangeNotSatisfiable) as exception:
            await send_http_start(
                send,
                exception.status_code,
                [
                    (k.encode("latin-1"), v.encode("latin-1"))
                    for k, v in (exception.headers or {}).items()
                ],
            )
            return await send_http_body(
                send,
                b"" if exception.content is None else exception.content.encode("utf8"),
            )

        if len(ranges) == 1:
            start, end = ranges[0]
            return await self.handle_single_range(
                send_header_only, file_size, scope, send, start, end
            )
        else:
            return await self.handle_several_ranges(
                send_header_only, file_size, scope, send, ranges
            )



from types import TracebackType
from typing import (
    Any,
    Awaitable,
    Callable,
    Iterable,
    List,
    MutableMapping,
    Optional,
    Tuple,
    Type,
    Final,
    Literal,
    Protocol,
    TypedDict,
    final,
    runtime_checkable,
)

__all__ = [
    "Scope",
    "Message",
    "Receive",
    "Send",
    "ASGIApp",
    "ExcInfo",
    "Environ",
    "StartResponse",
    "WSGIApp",
    # built-in types
    "TypedDict",
    "Literal",
    "Final",
    "final",
    "Protocol",
    "runtime_checkable",
]

# ASGI
Scope = MutableMapping[str, Any]

Message = MutableMapping[str, Any]

Receive = Callable[[], Awaitable[Message]]

Send = Callable[[Message], Awaitable[None]]


class ASGIApp(Protocol):
    def __call__(
        self, scope: Scope, receive: Receive, send: Send
    ) -> Awaitable[None]: ...


# WSGI: view PEP3333
Environ = MutableMapping[str, Any]

ExcInfo = Tuple[Type[BaseException], BaseException, Optional[TracebackType]]


class StartResponse(Protocol):
    def __call__(
        self,
        status: str,
        response_headers: List[Tuple[str, str]],
        exc_info: Optional[ExcInfo] = None,
    ) -> Any: ...


class WSGIApp(Protocol):
    def __call__(
        self, environ: Environ, start_response: StartResponse
    ) -> Iterable[bytes]: ...


# Server-sent Event
# https://developer.mozilla.org/en-US/docs/Web/API/Server-sent_events/Using_server-sent_events
ServerSentEvent = TypedDict(
    "ServerSentEvent", {"event": str, "data": str, "id": str, "retry": int}, total=False
)



import functools
from typing import Callable, Iterable

from baize.typing import Environ, StartResponse, WSGIApp

from .requests import Request
from .responses import Response

ViewType = Callable[[Request], Response]


def request_response(view: ViewType) -> WSGIApp:
    """
    This can turn a callable object into a WSGI application.

    ```python
    @request_response
    def f(request: Request) -> Response:
        ...
    ```
    """

    @functools.wraps(view)
    def wsgi(environ: Environ, start_response: StartResponse) -> Iterable[bytes]:
        request = Request(environ)
        response = view(request)
        yield from response(environ, start_response)

    return wsgi


def decorator(
    handler: Callable[[Request, ViewType], Response],
) -> Callable[[ViewType], ViewType]:
    """
    This can turn a callable object into a decorator for view.

    ```python
    @decorator
    def m(request: Request, next_call: Callable[[Request], Response]) -> Response:
        ...
        response = next_call(request)
        ...
        return response

    @request_response
    @m
    def v(request: Request) -> Response:
        ...
    ```
    """

    @functools.wraps(handler)
    def d(next_call: ViewType) -> ViewType:
        """
        This is the actual decorator.
        """

        @functools.wraps(next_call)
        def view(request: Request) -> Response:
            return handler(request, next_call)

        return view

    return d



from baize.routing import BaseHosts, BaseRouter, BaseSubpaths
from baize.typing import ASGIApp, Receive, Scope, Send

from .responses import PlainTextResponse, Response


class Router(BaseRouter[ASGIApp]):
    """
    A router to assign different paths to different ASGI applications.

    ```python
    applications = Router(
        ("/static/{filepath:any}", static_files),
        ("/api/{_:any}", api_app),
        ("/about/{name}", about_page),
        ("/", homepage),
    )
    """

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] == "lifespan":  # pragma: no cover
            raise RuntimeError("Unsupported lifespan in `Router`")
        result = self.search(scope["path"])
        if result is None:
            response: ASGIApp = Response(404)
        else:
            route, path_params = result
            scope["path_params"] = path_params
            response = route.endpoint
        return await response(scope, receive, send)


class Subpaths(BaseSubpaths[ASGIApp]):
    """
    A router allocates different prefix requests to different ASGI applications.

    NOTE: This will change the values of `scope["root_path"]` and `scope["path"]`.

    ```python
    applications = Subpaths(
        ("/static", static_files),
        ("/api", api_app),
        ("", default_app),
    )
    ```
    """

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] == "lifespan":  # pragma: no cover
            raise RuntimeError("Unsupported lifespan in `Subpaths`")
        path = scope["path"]
        result = self.search(path)
        if result is None:
            response: ASGIApp = Response(404)
        else:
            prefix, response = result
            scope["root_path"] = scope.get("root_path", "") + prefix
            scope["path"] = path[len(prefix) :]
        return await response(scope, receive, send)


class Hosts(BaseHosts[ASGIApp]):
    r"""
    A router that distributes requests to different ASGI applications based on Host.

    ```python
    applications = Hosts(
        (r"static\.example\.com", static_files),
        (r"api\.example\.com", api_app),
        (r"(www\.)?example\.com", default_app),
    )
    ```
    """

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] == "lifespan":  # pragma: no cover
            raise RuntimeError("Unsupported lifespan in `Hosts`")
        host = ""
        for k, v in scope["headers"]:
            if k == b"host":
                host = v.decode("latin-1")
        endpoint = self.search(host)
        if endpoint is None:
            response: ASGIApp = PlainTextResponse(b"Invalid host", 404)
        else:
            response = endpoint
        return await response(scope, receive, send)



import re
from typing import List, Optional, Tuple, Union, cast

from .datastructures import Headers
from .exceptions import MalformedMultipart
from .typing import Final
from .utils import parse_header

__all__ = [
    "Event",
    "Preamble",
    "Field",
    "File",
    "Data",
    "Epilogue",
    "NeedData",
    "NEED_DATA",
    "MultipartDecoder",
    "safe_decode",
]


class Event:
    __slots__ = ()


class Preamble(Event):
    __slots__ = ("data",)

    def __init__(self, data: bytes) -> None:
        self.data = data

    def __eq__(self, obj: object) -> bool:
        return isinstance(obj, self.__class__) and self.data == obj.data


class Field(Event):
    __slots__ = ("name", "headers")

    def __init__(self, name: str, headers: Headers) -> None:
        self.name = name
        self.headers = headers

    def __eq__(self, obj: object) -> bool:
        return (
            isinstance(obj, self.__class__)
            and self.name == obj.name
            and self.headers == obj.headers
        )


class File(Event):
    __slots__ = ("name", "filename", "headers")

    def __init__(self, name: str, filename: str, headers: Headers) -> None:
        self.name = name
        self.filename = filename
        self.headers = headers

    def __eq__(self, obj: object) -> bool:
        return (
            isinstance(obj, self.__class__)
            and self.name == obj.name
            and self.filename == obj.filename
            and self.headers == obj.headers
        )


class Data(Event):
    __slots__ = ("data", "more_data")

    def __init__(self, data: bytes, more_data: bool) -> None:
        self.data = data
        self.more_data = more_data

    def __eq__(self, obj: object) -> bool:
        return (
            isinstance(obj, self.__class__)
            and self.data == obj.data
            and self.more_data == obj.more_data
        )


class Epilogue(Event):
    __slots__ = ("data",)

    def __init__(self, data: bytes) -> None:
        self.data = data

    def __eq__(self, obj: object) -> bool:
        return isinstance(obj, self.__class__) and self.data == obj.data


class NeedData(Event):
    __slots__ = ()


NEED_DATA: Final = NeedData()


class State:
    PREAMBLE: Final = object()
    PART: Final = object()
    DATA: Final = object()
    EPILOGUE: Final = object()
    COMPLETE: Final = object()


# Multipart line breaks MUST be CRLF (\r\n) by RFC-7578, except that
# many implementations break this and either use CR or LF alone.
LINE_BREAK: Final = b"(?:\r\n|\n|\r)"
BLANK_LINE_RE: Final = re.compile(b"(?:\r\n\r\n|\r\r|\n\n)", re.MULTILINE)
LINE_BREAK_RE: Final = re.compile(LINE_BREAK, re.MULTILINE)
# Header values can be continued via a space or tab after the linebreak, as
# per RFC2231
HEADER_CONTINUATION_RE: Final = re.compile(b"%s[ \t]" % LINE_BREAK, re.MULTILINE)


class MultipartDecoder:
    """Decodes a multipart message as bytes into Python events.
    The part data is returned as available to allow the caller to save
    the data from memory to disk, if desired.

    https://github.com/pallets/werkzeug/blob/3f8a95fb275dcad27d541f00026b229636e1a485/src/werkzeug/sansio/multipart.py

    Copyright 2007 Pallets

    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions are
    met:

    1.  Redistributions of source code must retain the above copyright
        notice, this list of conditions and the following disclaimer.

    2.  Redistributions in binary form must reproduce the above copyright
        notice, this list of conditions and the following disclaimer in the
        documentation and/or other materials provided with the distribution.

    3.  Neither the name of the copyright holder nor the names of its
        contributors may be used to endorse or promote products derived from
        this software without specific prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
    PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
    HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
    SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
    TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
    PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
    LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
    NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
    SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
    """

    def __init__(self, boundary: bytes, charset: str) -> None:
        self.buffer = bytearray()
        self.complete = False
        self.state = State.PREAMBLE
        self.boundary = boundary
        self.charset = charset

        # Note in the below \h i.e. horizontal whitespace is used
        # as [^\S\n\r] as \h isn't supported in python.

        # The preamble must end with a boundary where the boundary is
        # prefixed by a line break, RFC2046. Except that many
        # implementations including Werkzeug's tests omit the line
        # break prefix. In addition the first boundary could be the
        # epilogue boundary (for empty form-data) hence the matching
        # group to understand if it is an epilogue boundary.
        self.preamble_re = re.compile(
            rb"%s?--%s(--[^\S\n\r]*%s?|[^\S\n\r]*%s)"
            % (LINE_BREAK, re.escape(boundary), LINE_BREAK, LINE_BREAK),
            re.MULTILINE,
        )
        # A boundary must include a line break prefix and suffix, and
        # may include trailing whitespace. In addition the boundary
        # could be the epilogue boundary hence the matching group to
        # understand if it is an epilogue boundary.
        self.boundary_re = re.compile(
            rb"%s--%s(--[^\S\n\r]*%s?|[^\S\n\r]*%s)"
            % (LINE_BREAK, re.escape(boundary), LINE_BREAK, LINE_BREAK),
            re.MULTILINE,
        )

    def last_newline(self) -> int:
        try:
            last_nl = self.buffer.rindex(b"\n")
        except ValueError:
            last_nl = len(self.buffer)
        try:
            last_cr = self.buffer.rindex(b"\r")
        except ValueError:
            last_cr = len(self.buffer)

        return min(last_nl, last_cr)

    def receive_data(self, data: Optional[bytes]) -> None:
        if data is None:
            self.complete = True
        else:
            self.buffer.extend(data)

    def next_event(self) -> Event:
        event: Event = NEED_DATA

        if self.state == State.PREAMBLE:
            match = self.preamble_re.search(self.buffer)
            if match is not None:
                if match.group(1).startswith(b"--"):
                    self.state = State.EPILOGUE
                else:
                    self.state = State.PART
                data = bytes(self.buffer[: match.start()])
                del self.buffer[: match.end()]
                event = Preamble(data=data)

        elif self.state == State.PART:
            match = BLANK_LINE_RE.search(self.buffer)
            if match is not None:
                headers = self._parse_headers(self.buffer[: match.start()])
                del self.buffer[: match.end()]

                if "content-disposition" not in headers:  # pragma: no cover
                    raise MalformedMultipart("Missing Content-Disposition header")

                disposition, extra = parse_header(headers["content-disposition"])
                name = cast(str, extra.get("name"))
                filename = extra.get("filename")
                if filename is not None:
                    event = File(filename=filename, headers=headers, name=name)
                else:
                    event = Field(headers=headers, name=name)
                self.state = State.DATA

        elif self.state == State.DATA:
            if self.buffer.find(b"--" + self.boundary) == -1:
                # No complete boundary in the buffer, but there may be
                # a partial boundary at the end. As the boundary
                # starts with either a nl or cr find the earliest and
                # return up to that as data.
                data_length = del_index = self.last_newline()
                more_data = True
            else:
                match = self.boundary_re.search(self.buffer)
                if match is not None:
                    if match.group(1).startswith(b"--"):
                        self.state = State.EPILOGUE
                    else:
                        self.state = State.PART
                    data_length = match.start()
                    del_index = match.end()
                else:
                    data_length = del_index = self.last_newline()
                more_data = match is None

            data = bytes(self.buffer[:data_length])
            del self.buffer[:del_index]
            if data or not more_data:
                event = Data(data=data, more_data=more_data)

        elif self.state == State.EPILOGUE and self.complete:
            event = Epilogue(data=bytes(self.buffer))
            del self.buffer[:]
            self.state = State.COMPLETE

        if self.complete and isinstance(event, NeedData):  # pragma: no cover
            raise MalformedMultipart(
                f"Invalid form-data cannot parse beyond {self.state}"
            )

        return event

    def _parse_headers(self, data: bytes) -> Headers:
        headers: List[Tuple[str, str]] = []
        # Merge the continued headers into one line
        data = HEADER_CONTINUATION_RE.sub(b" ", data)
        # Now there is one header per line
        for line in data.splitlines():
            line = line.strip()
            if line != b"":
                name, value = safe_decode(line, self.charset).split(":", 1)
                headers.append((name.strip(), value.strip()))
        return Headers(headers)


def safe_decode(src: Union[bytes, bytearray], charset: str) -> str:
    try:
        return src.decode(charset)
    except (UnicodeDecodeError, LookupError):
        return src.decode("latin-1")



from baize.asgi.helper import empty_receive, empty_send, send_http_body, send_http_start
from baize.asgi.requests import ClientDisconnect, HTTPConnection, Request
from baize.asgi.responses import (
    FileResponse,
    HTMLResponse,
    JSONResponse,
    PlainTextResponse,
    RedirectResponse,
    Response,
    SendEventResponse,
    SmallResponse,
    StreamResponse,
)
from baize.asgi.routing import Hosts, Router, Subpaths
from baize.asgi.shortcut import decorator, request_response, websocket_session
from baize.asgi.staticfiles import Files, Pages
from baize.asgi.websocket import (
    WebSocket,
    WebsocketDenialResponse,
    WebSocketDisconnect,
    WebSocketState,
)
from baize.asgi.middleware import NextRequest, NextResponse, middleware, PipeStream

__all__ = [
    "empty_receive",
    "empty_send",
    "send_http_start",
    "send_http_body",
    "ClientDisconnect",
    "HTTPConnection",
    "Request",
    "WebSocket",
    "WebSocketDisconnect",
    "WebsocketDenialResponse",
    "WebSocketState",
    "Response",
    "SmallResponse",
    "PlainTextResponse",
    "HTMLResponse",
    "JSONResponse",
    "RedirectResponse",
    "StreamResponse",
    "FileResponse",
    "SendEventResponse",
    "Router",
    "Subpaths",
    "Hosts",
    "Files",
    "Pages",
    "request_response",
    "websocket_session",
    "decorator",
    "NextRequest",
    "NextResponse",
    "middleware",
    "PipeStream",
]



import os
import stat
from typing import Iterable, Optional

from baize import staticfiles
from baize.datastructures import URL
from baize.exceptions import HTTPException
from baize.typing import Environ, StartResponse, WSGIApp

from .responses import FileResponse, RedirectResponse, Response


class Files(staticfiles.BaseFiles[WSGIApp]):
    """
    Provide the WSGI application to download files in the specified path or
    the specified directory under the specified package.

    Support request range and cache (304 status code).
    """

    def file_response(
        self,
        filepath: str,
        stat_result: os.stat_result,
        if_none_match: str,
        if_modified_since: str,
    ) -> Response:
        if self.if_none_match(
            FileResponse.generate_etag(stat_result), if_none_match
        ) or self.if_modified_since(stat_result.st_mtime, if_modified_since):
            response = Response(304)
        else:
            response = FileResponse(filepath, stat_result=stat_result)
        self.set_response_headers(response)
        return response

    def __call__(
        self, environ: Environ, start_response: StartResponse
    ) -> Iterable[bytes]:
        if_none_match: str = environ.get("HTTP_IF_NONE_MATCH", "")
        if_modified_since: str = environ.get("HTTP_IF_MODIFIED_SINCE", "")
        filepath = self.ensure_absolute_path(environ.get("PATH_INFO", ""))
        stat_result, is_file = self.check_path_is_file(filepath)
        if is_file and stat_result:
            assert filepath is not None  # Just for type check
            return self.file_response(
                filepath, stat_result, if_none_match, if_modified_since
            )(environ, start_response)

        if self.handle_404 is None:
            raise HTTPException(404)
        else:
            return self.handle_404(environ, start_response)


class Pages(Files):
    """
    Provide the WSGI application to download files in the specified path or
    the specified directory under the specified package.
    Unlike `Files`, when you visit a directory, it will try to return the content
    of the file named `index.html` in that directory. Or if the `pathname.html` is
    exist, it will return the content of that file.

    Support request range and cache (304 status code).
    """

    def __call__(
        self, environ: Environ, start_response: StartResponse
    ) -> Iterable[bytes]:
        if_none_match: str = environ.get("HTTP_IF_NONE_MATCH", "")
        if_modified_since: str = environ.get("HTTP_IF_MODIFIED_SINCE", "")
        filepath = self.ensure_absolute_path(environ.get("PATH_INFO", ""))
        stat_result, is_file = self.check_path_is_file(filepath)
        if (
            stat_result is None  # filepath is not exist
            and filepath is not None  # Just for type check
            and not filepath.endswith(".html")  # filepath is not a html file
        ):
            filepath += ".html"
            stat_result, is_file = self.check_path_is_file(filepath)

        if stat_result is not None:
            assert filepath is not None  # Just for type check
            if is_file:
                return self.file_response(
                    filepath, stat_result, if_none_match, if_modified_since
                )(environ, start_response)
            if stat.S_ISDIR(stat_result.st_mode):
                url = URL(environ=environ)
                url = url.replace(scheme="", path=url.path + "/")
                return RedirectResponse(url)(environ, start_response)

        if self.handle_404 is None:
            raise HTTPException(404)
        else:
            return self.handle_404(environ, start_response)

    def ensure_absolute_path(self, path: str) -> Optional[str]:
        abspath = super().ensure_absolute_path(path)
        if abspath is not None:
            if abspath.endswith("/"):
                abspath += "index.html"
        return abspath



from typing import Iterable

from baize.routing import BaseHosts, BaseRouter, BaseSubpaths
from baize.typing import Environ, StartResponse, WSGIApp

from .responses import PlainTextResponse, Response


class Router(BaseRouter[WSGIApp]):
    """
    A router to assign different paths to different WSGI applications.

    ```python
    applications = Router(
        ("/static/{filepath:any}", static_files),
        ("/api/{_:any}", api_app),
        ("/about/{name}", about_page),
        ("/", homepage),
    )
    ```
    """

    def __call__(
        self, environ: Environ, start_response: StartResponse
    ) -> Iterable[bytes]:
        result = self.search(environ.get("PATH_INFO", ""))
        if result is None:
            response: WSGIApp = Response(404)
        else:
            route, path_params = result
            environ["PATH_PARAMS"] = path_params
            response = route.endpoint
        yield from response(environ, start_response)


class Subpaths(BaseSubpaths[WSGIApp]):
    """
    A router allocates different prefix requests to different WSGI applications.

    NOTE: This will change the values of `environ["SCRIPT_NAME"]` and `environ["PATH_INFO"]`.

    ```python
    applications = Subpaths(
        ("/static", static_files),
        ("/api", api_app),
        ("", default_app),
    )
    ```
    """

    def __call__(
        self, environ: Environ, start_response: StartResponse
    ) -> Iterable[bytes]:
        path = environ.get("PATH_INFO", "")
        result = self.search(path)
        if result is None:
            response: WSGIApp = Response(404)
        else:
            prefix, response = result
            environ["SCRIPT_NAME"] = environ.get("SCRIPT_NAME", "") + prefix
            environ["PATH_INFO"] = path[len(prefix) :]
        yield from response(environ, start_response)


class Hosts(BaseHosts[WSGIApp]):
    r"""
    A router that distributes requests to different WSGI applications based on Host.

    ```python
    applications = Hosts(
        (r"static\.example\.com", static_files),
        (r"api\.example\.com", api_app),
        (r"(www\.)?example\.com", default_app),
    )
    ```
    """

    def __call__(
        self, environ: Environ, start_response: StartResponse
    ) -> Iterable[bytes]:
        endpoint = self.search(environ.get("HTTP_HOST", ""))
        if endpoint is None:
            response: WSGIApp = PlainTextResponse(b"Invalid host", 404)
        else:
            response = endpoint
        yield from response(environ, start_response)



from typing import Iterable, Optional, Tuple

from baize.typing import Message, Send


async def send_http_start(
    send: Send,
    status_code: int,
    headers: Optional[Iterable[Tuple[bytes, bytes]]] = None,
) -> None:
    """
    helper function for send http.response.start

    https://asgi.readthedocs.io/en/latest/specs/www.html#response-start-send-event
    """
    message = {"type": "http.response.start", "status": status_code}
    if headers is not None:
        message["headers"] = headers
    await send(message)


async def send_http_body(
    send: Send, body: bytes = b"", *, more_body: bool = False
) -> None:
    """
    helper function for send http.response.body

    https://asgi.readthedocs.io/en/latest/specs/www.html#response-body-send-event
    """
    await send({"type": "http.response.body", "body": body, "more_body": more_body})


async def empty_receive() -> Message:
    raise NotImplementedError("Receive channel has not been made available")


async def empty_send(message: Message) -> None:
    raise NotImplementedError("Send channel has not been made available")



import asyncio
import functools
import inspect
import typing

T = typing.TypeVar("T")


class cached_property(typing.Generic[T]):
    """
    A property that is only computed once per instance and then replaces
    itself with an ordinary attribute. Deleting the attribute resets the
    property.
    """

    def __init__(self, func: typing.Callable[..., T]) -> None:
        self.func = func
        functools.update_wrapper(self, func)  # type: ignore

    @typing.overload
    def __get__(self, obj: None, cls: type) -> "cached_property": ...

    @typing.overload
    def __get__(self, obj: object, cls: type) -> T: ...

    def __get__(self, obj, cls):
        if obj is None:
            value = self
        else:
            result = self.func(obj)
            if inspect.isawaitable(result):
                result = asyncio.ensure_future(result)
            value = obj.__dict__[self.func.__name__] = result
        return value


# ############################################################################
# ############################# COPY From stdlib #############################
# ############ Because https://peps.python.org/pep-0594/ #####################
# ############################################################################


def _parseparam(s: str) -> typing.Generator[str, None, None]:
    while s[:1] == ";":
        s = s[1:]
        end = s.find(";")
        while end > 0 and (s.count('"', 0, end) - s.count('\\"', 0, end)) % 2:
            end = s.find(";", end + 1)
        if end < 0:
            end = len(s)
        f = s[:end]
        yield f.strip()
        s = s[end:]


def parse_header(line: str) -> typing.Tuple[str, typing.Dict[str, str]]:
    """Parse a Content-type like header.

    Return the main content-type and a dictionary of options.

    """
    parts = _parseparam(";" + line)
    key = parts.__next__()
    pdict = {}
    for p in parts:
        i = p.find("=")
        if i >= 0:
            name = p[:i].strip().lower()
            value = p[i + 1 :].strip()
            if len(value) >= 2 and value[0] == value[-1] == '"':
                value = value[1:-1]
                value = value.replace("\\\\", "\\").replace('\\"', '"')
            pdict[name] = value
    return key, pdict



from http import HTTPStatus
from typing import Any, Generic, Mapping, NoReturn, Optional, TypeVar

T = TypeVar("T")


class HTTPException(Exception, Generic[T]):
    """
    Base HTTP Exception
    """

    def __init__(
        self,
        status_code: int = 400,
        headers: Optional[Mapping[str, str]] = None,
        content: Optional[T] = None,
    ) -> None:
        self.status_code = status_code
        self.headers = headers
        self.content = content
        if content is not None:
            status_description = repr(content)
        else:
            try:
                status_description = HTTPStatus(status_code).description
            except ValueError:
                status_description = "Maybe a custom HTTP status code"
        super().__init__(status_code, status_description)


def abort(
    status_code: int = 400,
    headers: Optional[Mapping[str, str]] = None,
    content: Any = None,
) -> NoReturn:
    """
    raise a `HTTPException`. Parameters are completely consistent with `HTTPException`.
    """
    raise HTTPException(status_code=status_code, headers=headers, content=content)


class RequestEntityTooLarge(HTTPException[None]):
    """
    413 Request Entity Too Large
    """

    def __init__(self, retry_after: Optional[int] = None) -> None:
        super().__init__(
            413, {"Retry-After": str(retry_after)} if retry_after is not None else None
        )


class UnsupportedMediaType(HTTPException[None]):
    """
    415 Unsupported Media Type
    """

    def __init__(self, *supported_media_types: str) -> None:
        super().__init__(415, {"Accpet": ", ".join(supported_media_types)}, None)


class RangeNotSatisfiable(HTTPException[None]):
    """
    416 Range Not Satisfiable
    """

    def __init__(self, max_size: int) -> None:
        super().__init__(416, {"Content-Range": f"*/{max_size}"}, None)


# ###################################################################################
# ################################ Custom Exception #################################
# ###################################################################################


class MalformedJSON(HTTPException[str]):
    def __init__(self, message: str = "Malformed JSON") -> None:
        super().__init__(content=message)


class MalformedMultipart(HTTPException[str]):
    def __init__(self, message: str = "Malformed multipart") -> None:
        super().__init__(content=message)


class MalformedRangeHeader(HTTPException[str]):
    def __init__(self, message: str = "Malformed Range header") -> None:
        super().__init__(content=message)



import functools
from typing import Awaitable, Callable

from baize.typing import ASGIApp, Receive, Scope, Send

from .requests import Request
from .responses import Response
from .websocket import WebSocket, WebsocketDenialResponse

ViewType = Callable[[Request], Awaitable[Response]]


def request_response(view: ViewType) -> ASGIApp:
    """
    This can turn a callable object into a ASGI application.

    ```python
    @request_response
    async def f(request: Request) -> Response:
        ...
    ```
    """

    @functools.wraps(view)
    async def asgi(scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] == "websocket":
            return await WebsocketDenialResponse(Response(404))(scope, receive, send)
        else:
            request = Request(scope, receive, send)
            resposne = await view(request)
            return await resposne(scope, receive, send)

    return asgi


def websocket_session(view: Callable[[WebSocket], Awaitable[None]]) -> ASGIApp:
    """
    This can turn a callable object into a ASGI application.

    ```python
    @websocket_session
    async def f(websocket: WebSocket) -> None:
        ...
    ```
    """

    @functools.wraps(view)
    async def asgi(scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] == "http":
            await Response(404)(scope, receive, send)
        else:
            websocket = WebSocket(scope, receive, send)
            await view(websocket)

    return asgi


def decorator(
    handler: Callable[[Request, ViewType], Awaitable[Response]],
) -> Callable[[ViewType], ViewType]:
    """
    This can turn a callable object into a decorator for view.

    ```python
    @decorator
    async def m(request: Request, next_call: Callable[[Request], Awaitable[Response]]) -> Response:
        ...
        response = await next_call(request)
        ...
        return response


    @request_response
    @m
    async def v(request: Request) -> Response:
        ...
    ```
    """

    @functools.wraps(handler)
    def d(next_call: ViewType) -> ViewType:
        """
        This is the actual decorator.
        """

        @functools.wraps(next_call)
        async def view(request: Request) -> Response:
            return await handler(request, next_call)

        return view

    return d
